#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from functools import reduce
import numpy as np
from pyscf import lib
from pyscf.lib import logger
from pyscf.tdscf import uhf
from pyscf.tdscf._lr_eig import eigh as lr_eigh, eig as lr_eig
from pyscf.pbc import scf
from pyscf.pbc.tdscf.krhf import KTDBase, _get_e_ia
from pyscf.pbc.lib.kpts_helper import is_gamma_point, get_kconserv_ria, conj_mapping
from pyscf.pbc.scf import _response_functions  # noqa
from pyscf import __config__

REAL_EIG_THRESHOLD = getattr(__config__, 'pbc_tdscf_uhf_TDDFT_pick_eig_threshold', 1e-3)

def get_ab(mf, kshift=0):
    r'''A and B matrices for TDDFT response function.

    A[i,a,j,b] = \delta_{ab}\delta_{ij}(E_a - E_i) + (ai||jb)
    B[i,a,j,b] = (ai||bj)

    Ref: Chem Phys Lett, 256, 454

    Kwargs:
        kshift : integer
            The index of the k-point that represents the transition between
            k-points in the excitation coefficients.
    '''
    cell = mf.cell
    mo_energy = scf.addons.mo_energy_with_exxdiv_none(mf)
    mo_a, mo_b = mo = np.asarray(mf.mo_coeff)
    mo_occ = np.asarray(mf.mo_occ)
    kpts = mf.kpts
    nkpts, nao, nmo = mo_a.shape
    noccs = np.count_nonzero(mo_occ!=0, axis=2)
    nocc = noccs[0,0]
    nvir = nmo - nocc
    assert np.all(noccs == nocc)
    nocc_a = nocc_b = nocc
    nvir_a = nvir_b = nvir
    orbo_a = mo_a[:,:,:nocc]
    orbo_b = mo_b[:,:,:nocc]
    orbv_a = mo_a[:,:,nocc:]
    orbv_b = mo_b[:,:,nocc:]

    kconserv = get_kconserv_ria(cell, kpts)[kshift]
    e_ia_a = np.asarray(_get_e_ia(mo_energy[0], mo_occ[0], kconserv)).astype(mo.dtype)
    e_ia_b = np.asarray(_get_e_ia(mo_energy[1], mo_occ[1], kconserv)).astype(mo.dtype)
    a_aa = np.diag(e_ia_a.ravel()).reshape(nkpts,nocc_a,nvir_a,nkpts,nocc_a,nvir_a)
    a_bb = np.diag(e_ia_b.ravel()).reshape(nkpts,nocc_b,nvir_b,nkpts,nocc_b,nvir_b)
    a_ab = np.zeros((nkpts,nocc_a,nvir_a,nkpts,nocc_b,nvir_b), dtype=a_aa.dtype)
    b_aa = np.zeros_like(a_aa)
    b_bb = np.zeros_like(a_bb)
    b_ab = np.zeros_like(a_ab)
    a = (a_aa, a_ab, a_bb)
    b = (b_aa, b_ab, b_bb)
    weight = 1./nkpts

    def add_hf_(a, b, hyb=1):
        eri_aa = mf.with_df.ao2mo_7d([mo_a,orbo_a,mo_a,mo_a], kpts)
        eri_ab = mf.with_df.ao2mo_7d([mo_a,orbo_a,mo_b,mo_b], kpts)
        eri_bb = mf.with_df.ao2mo_7d([mo_b,orbo_b,mo_b,mo_b], kpts)
        eri_aa *= weight
        eri_ab *= weight
        eri_bb *= weight
        eri_aa.reshape(nkpts,nkpts,nkpts,nmo,nocc_a,nmo,nmo)
        eri_ab.reshape(nkpts,nkpts,nkpts,nmo,nocc_a,nmo,nmo)
        eri_bb.reshape(nkpts,nkpts,nkpts,nmo,nocc_b,nmo,nmo)
        a_aa, a_ab, a_bb = a
        b_aa, b_ab, b_bb = b

        for ki, ka in enumerate(kconserv):
            for kj, kb in enumerate(kconserv):
                a_aa[ki,:,:,kj] += np.einsum('aijb->iajb', eri_aa[ka,ki,kj,nocc_a:,:,:nocc_a,nocc_a:])
                a_aa[ki,:,:,kj] -= np.einsum('jiab->iajb', eri_aa[kj,ki,ka,:nocc_a,:,nocc_a:,nocc_a:]) * hyb
                a_bb[ki,:,:,kj] += np.einsum('aijb->iajb', eri_bb[ka,ki,kj,nocc_b:,:,:nocc_b,nocc_b:])
                a_bb[ki,:,:,kj] -= np.einsum('jiab->iajb', eri_bb[kj,ki,ka,:nocc_b,:,nocc_b:,nocc_b:]) * hyb
                a_ab[ki,:,:,kj] += np.einsum('aijb->iajb', eri_ab[ka,ki,kj,nocc_a:,:,:nocc_b,nocc_b:])

            for kb, kj in enumerate(kconserv):
                b_aa[ki,:,:,kj] += np.einsum('aibj->iajb', eri_aa[ka,ki,kb,nocc_a:,:,nocc_a:,:nocc_a])
                b_aa[ki,:,:,kj] -= np.einsum('ajbi->iajb', eri_aa[ka,kj,kb,nocc_a:,:,nocc_a:,:nocc_a]) * hyb
                b_bb[ki,:,:,kj] += np.einsum('aibj->iajb', eri_bb[ka,ki,kb,nocc_b:,:,nocc_b:,:nocc_b])
                b_bb[ki,:,:,kj] -= np.einsum('ajbi->iajb', eri_bb[ka,kj,kb,nocc_b:,:,nocc_b:,:nocc_b]) * hyb
                b_ab[ki,:,:,kj] += np.einsum('aibj->iajb', eri_ab[ka,ki,kb,nocc_a:,:,nocc_b:,:nocc_b])

    if isinstance(mf, scf.hf.KohnShamDFT):
        ni = mf._numint
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, cell.spin)

        add_hf_(a, b, hyb)
        if omega != 0:  # For RSH
            raise NotImplementedError

        xctype = ni._xc_type(mf.xc)
        dm0 = mf.make_rdm1(mo, mo_occ)
        make_rho = ni._gen_rho_evaluator(cell, dm0, hermi=1, with_lapl=False)[0]
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)
        cmap = conj_mapping(cell, kpts)

        if xctype == 'LDA':
            ao_deriv = 0
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho0a = make_rho(0, ao, mask, xctype)
                rho0b = make_rho(1, ao, mask, xctype)
                rho = (rho0a, rho0b)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc[0,0] * weight

                rho_o_a = lib.einsum('krp,kpi->kri', ao, orbo_a)
                rho_v_a = lib.einsum('krp,kpi->kri', ao, orbv_a)
                rho_o_b = lib.einsum('krp,kpi->kri', ao, orbo_b)
                rho_v_b = lib.einsum('krp,kpi->kri', ao, orbv_b)
                rho_ov_a = np.einsum('kri,kra->kria', rho_o_a, rho_v_a)
                rho_ov_b = np.einsum('kri,kra->kria', rho_o_b, rho_v_b)
                rho_vo_a = rho_ov_a.conj()[cmap]
                rho_vo_b = rho_ov_b.conj()[cmap]
                w_vo_aa = np.einsum('kria,r->kria', rho_vo_a, wfxc[0,0]) * (1/nkpts)
                w_vo_ab = np.einsum('kria,r->kria', rho_vo_a, wfxc[0,1]) * (1/nkpts)
                w_vo_bb = np.einsum('kria,r->kria', rho_vo_b, wfxc[1,1]) * (1/nkpts)

                a_aa += lib.einsum('kria,lrjb->kialjb', w_vo_aa, rho_ov_a)
                b_aa += lib.einsum('kria,lrjb->kialjb', w_vo_aa, rho_vo_a)

                a_ab += lib.einsum('kria,lrjb->kialjb', w_vo_ab, rho_ov_b)
                b_ab += lib.einsum('kria,lrjb->kialjb', w_vo_ab, rho_vo_b)

                a_bb += lib.einsum('kria,lrjb->kialjb', w_vo_bb, rho_ov_b)
                b_bb += lib.einsum('kria,lrjb->kialjb', w_vo_bb, rho_vo_b)

        elif xctype == 'GGA':
            ao_deriv = 1
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho0a = make_rho(0, ao, mask, xctype)
                rho0b = make_rho(1, ao, mask, xctype)
                rho = (rho0a, rho0b)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc * weight

                rho_o_a = lib.einsum('kxrp,kpi->kxri', ao, orbo_a)
                rho_v_a = lib.einsum('kxrp,kpi->kxri', ao, orbv_a)
                rho_o_b = lib.einsum('kxrp,kpi->kxri', ao, orbo_b)
                rho_v_b = lib.einsum('kxrp,kpi->kxri', ao, orbv_b)
                rho_ov_a = np.einsum('kxri,kra->kxria', rho_o_a, rho_v_a[:,0])
                rho_ov_b = np.einsum('kxri,kra->kxria', rho_o_b, rho_v_b[:,0])
                rho_ov_a[:,1:4] += np.einsum('kri,kxra->kxria', rho_o_a[:,0], rho_v_a[:,1:4])
                rho_ov_b[:,1:4] += np.einsum('kri,kxra->kxria', rho_o_b[:,0], rho_v_b[:,1:4])
                rho_vo_a = rho_ov_a.conj()[cmap]
                rho_vo_b = rho_ov_b.conj()[cmap]
                w_vo_aa = np.einsum('xyr,kxria->kyria', wfxc[0,:,0], rho_vo_a) * (1/nkpts)
                w_vo_ab = np.einsum('xyr,kxria->kyria', wfxc[0,:,1], rho_vo_a) * (1/nkpts)
                w_vo_bb = np.einsum('xyr,kxria->kyria', wfxc[1,:,1], rho_vo_b) * (1/nkpts)

                a_aa += lib.einsum('kxria,lxrjb->kialjb', w_vo_aa, rho_ov_a)
                b_aa += lib.einsum('kxria,lxrjb->kialjb', w_vo_aa, rho_vo_a)

                a_ab += lib.einsum('kxria,lxrjb->kialjb', w_vo_ab, rho_ov_b)
                b_ab += lib.einsum('kxria,lxrjb->kialjb', w_vo_ab, rho_vo_b)

                a_bb += lib.einsum('kxria,lxrjb->kialjb', w_vo_bb, rho_ov_b)
                b_bb += lib.einsum('kxria,lxrjb->kialjb', w_vo_bb, rho_vo_b)

        elif xctype == 'HF':
            pass

        elif xctype == 'NLC':
            raise NotImplementedError('NLC')

        elif xctype == 'MGGA':
            ao_deriv = 1
            for ao, _, mask, weight, coords \
                    in ni.block_loop(cell, mf.grids, nao, ao_deriv, kpts, None, max_memory):
                rho0a = make_rho(0, ao, mask, xctype)
                rho0b = make_rho(1, ao, mask, xctype)
                rho = (rho0a, rho0b)
                fxc = ni.eval_xc_eff(mf.xc, rho, deriv=2, xctype=xctype)[2]
                wfxc = fxc * weight

                rho_o_a = lib.einsum('kxrp,kpi->kxri', ao, orbo_a)
                rho_o_b = lib.einsum('kxrp,kpi->kxri', ao, orbo_b)
                rho_v_a = lib.einsum('kxrp,kpi->kxri', ao, orbv_a)
                rho_v_b = lib.einsum('kxrp,kpi->kxri', ao, orbv_b)
                rho_ov_a = np.einsum('kxri,kra->kxria', rho_o_a, rho_v_a[:,0])
                rho_ov_b = np.einsum('kxri,kra->kxria', rho_o_b, rho_v_b[:,0])
                rho_ov_a[:,1:4] += np.einsum('ri,xra->xria', rho_o_a[:,0], rho_v_a[:,1:4])
                rho_ov_b[:,1:4] += np.einsum('ri,xra->xria', rho_o_b[:,0], rho_v_b[:,1:4])
                tau_ov_a = np.einsum('kxri,kxra->kria', rho_o_a[:,1:4], rho_v_a[:,1:4]) * .5
                tau_ov_b = np.einsum('kxri,kxra->kria', rho_o_b[:,1:4], rho_v_b[:,1:4]) * .5
                rho_ov_a = np.vstack([rho_ov_a, tau_ov_a[:,np.newaxis]])
                rho_ov_b = np.vstack([rho_ov_b, tau_ov_b[:,np.newaxis]])
                rho_vo_a = rho_ov_a.conj()[cmap]
                rho_vo_b = rho_ov_b.conj()[cmap]
                w_vo_aa = np.einsum('xyr,kxria->kyria', wfxc[0,:,0], rho_vo_a) * (1/nkpts)
                w_vo_ab = np.einsum('xyr,kxria->kyria', wfxc[0,:,1], rho_vo_a) * (1/nkpts)
                w_vo_bb = np.einsum('xyr,kxria->kyria', wfxc[1,:,1], rho_vo_b) * (1/nkpts)

                a_aa += lib.einsum('kxria,lxrjb->kilajb', w_vo_aa, rho_ov_a)
                b_aa += lib.einsum('kxria,lxrjb->kilajb', w_vo_aa, rho_vo_a)

                a_ab += lib.einsum('kxria,lxrjb->kilajb', w_vo_ab, rho_ov_b)
                b_ab += lib.einsum('kxria,lxrjb->kilajb', w_vo_ab, rho_vo_b)

                a_bb += lib.einsum('kxria,lxrjb->kilajb', w_vo_bb, rho_ov_b)
                b_bb += lib.einsum('kxria,lxrjb->kilajb', w_vo_bb, rho_vo_b)
    else:
        add_hf_(a, b)

    return a, b

class TDA(KTDBase):

    def get_ab(self, mf=None, kshift=0):
        if mf is None: mf = self._scf
        return get_ab(mf, kshift)

    def gen_vind(self, mf, kshift=0):
        '''Compute Ax

        Kwargs:
            kshift : integer
                The index of the k-point that represents the transition between
                k-points in the excitation coefficients.
        '''
        assert mf is self._scf
        kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        nkpts = len(mo_occ[0])
        nao, nmo = mo_coeff[0][0].shape
        occidxa = [mo_occ[0][k]> 0 for k in range(nkpts)]
        occidxb = [mo_occ[1][k]> 0 for k in range(nkpts)]
        viridxa = [mo_occ[0][k]==0 for k in range(nkpts)]
        viridxb = [mo_occ[1][k]==0 for k in range(nkpts)]
        orboa = [mo_coeff[0][k][:,occidxa[k]] for k in range(nkpts)]
        orbob = [mo_coeff[1][k][:,occidxb[k]] for k in range(nkpts)]
        orbva = [mo_coeff[0][k][:,viridxa[k]] for k in range(nkpts)]
        orbvb = [mo_coeff[1][k][:,viridxb[k]] for k in range(nkpts)]
        dtype = np.result_type(*mo_coeff[0])

        moe = scf.addons.mo_energy_with_exxdiv_none(mf)
        e_ia_a = _get_e_ia(moe[0], mo_occ[0], kconserv)
        e_ia_b = _get_e_ia(moe[1], mo_occ[1], kconserv)
        hdiag = np.hstack([x.ravel() for x in (e_ia_a + e_ia_b)])

        mem_now = lib.current_memory()[0]
        max_memory = max(2000, self.max_memory*.8-mem_now)
        vresp = self.gen_response(hermi=0, max_memory=max_memory)

        def vind(zs):
            nz = len(zs)
            zs = [_unpack(z, mo_occ, kconserv) for z in zs]
            dms = np.empty((2,nz,nkpts,nao,nao), dtype=dtype)
            for i in range(nz):
                dm1a, dm1b = zs[i]
                for k, kp in enumerate(kconserv):
                    dms[0,i,kp] = lib.einsum('ov,pv,qo->pq', dm1a[k], orbva[kp], orboa[k].conj())
                    dms[1,i,kp] = lib.einsum('ov,pv,qo->pq', dm1b[k], orbvb[kp], orbob[k].conj())

            with lib.temporary_env(mf, exxdiv=None):
                v1ao = vresp(dms, kshift)
                v1ao = v1ao.reshape(2,nz,nkpts,nao,nao)

            v1s = []
            for i in range(nz):
                dm1a, dm1b = zs[i]
                v1as = [None] * nkpts
                v1bs = [None] * nkpts
                for k, kp in enumerate(kconserv):
                    v1a = lib.einsum('pq,qo,pv->ov', v1ao[0,i,kp], orboa[k], orbva[kp].conj())
                    v1b = lib.einsum('pq,qo,pv->ov', v1ao[1,i,kp], orbob[k], orbvb[kp].conj())
                    v1a += e_ia_a[k] * dm1a[k]
                    v1b += e_ia_b[k] * dm1b[k]
                    v1as[k] = v1a.ravel()
                    v1bs[k] = v1b.ravel()
                v1s.append( np.concatenate(v1as + v1bs) )
            return np.stack(v1s)

        return vind, hdiag

    def get_init_guess(self, mf, kshift, nstates=None):
        if nstates is None: nstates = self.nstates

        mo_energy = mf.mo_energy
        mo_occ = mf.mo_occ
        kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]
        e_ia_a = _get_e_ia(mo_energy[0], mo_occ[0], kconserv)
        e_ia_b = _get_e_ia(mo_energy[1], mo_occ[1], kconserv)
        e_ia = np.hstack([x.ravel() for x in (e_ia_a + e_ia_b)])

        nov = e_ia.size
        nstates = min(nstates, nov)
        e_threshold = np.partition(e_ia, nstates-1)[nstates-1]
        e_threshold += self.deg_eia_thresh

        idx = np.where(e_ia <= e_threshold)[0]
        x0 = np.zeros((idx.size, nov))
        for i, j in enumerate(idx):
            x0[i, j] = 1  # Koopmans' excitations
        return x0

    def kernel(self, x0=None):
        '''TDA diagonalization solver
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()

        log = logger.new_logger(self)

        mf = self._scf
        mo_occ = mf.mo_occ

        def pickeig(w, v, nroots, envs):
            idx = np.where(w > self.positive_eig_threshold)[0]
            return w[idx], v[:,idx], idx

        log = logger.Logger(self.stdout, self.verbose)

        self.converged = []
        self.e = []
        self.xy = []
        for i,kshift in enumerate(self.kshift_lst):
            kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

            vind, hdiag = self.gen_vind(self._scf, kshift)
            precond = self.get_precond(hdiag)

            if x0 is None:
                x0k = self.get_init_guess(self._scf, kshift, self.nstates)
            else:
                x0k = x0[i]

            converged, e, x1 = lr_eigh(
                vind, x0k, precond, tol_residual=self.conv_tol, lindep=self.lindep,
                nroots=self.nstates, pick=pickeig, max_cycle=self.max_cycle,
                max_memory=self.max_memory, verbose=log)
            self.converged.append( converged )
            self.e.append( e )
            self.xy.append( [(_unpack(xi, mo_occ, kconserv),  # (X_alpha, X_beta)
                        (0, 0))  # (Y_alpha, Y_beta)
                       for xi in x1] )
        #TODO: analyze CIS wfn point group symmetry
        log.timer(self.__class__.__name__, *cpu0)
        self._finalize()
        return self.e, self.xy
CIS = KTDA = TDA


class TDHF(KTDBase):

    get_ab = TDA.get_ab

    def gen_vind(self, mf, kshift=0):
        assert mf is self._scf
        assert kshift == 0

        mo_coeff = mf.mo_coeff
        mo_occ = mf.mo_occ
        nkpts = len(mo_occ[0])
        nao, nmo = mo_coeff[0][0].shape
        occidxa = [mo_occ[0][k]> 0 for k in range(nkpts)]
        occidxb = [mo_occ[1][k]> 0 for k in range(nkpts)]
        viridxa = [mo_occ[0][k]==0 for k in range(nkpts)]
        viridxb = [mo_occ[1][k]==0 for k in range(nkpts)]
        orboa = [mo_coeff[0][k][:,occidxa[k]] for k in range(nkpts)]
        orbob = [mo_coeff[1][k][:,occidxb[k]] for k in range(nkpts)]
        orbva = [mo_coeff[0][k][:,viridxa[k]] for k in range(nkpts)]
        orbvb = [mo_coeff[1][k][:,viridxb[k]] for k in range(nkpts)]
        dtype = np.result_type(*mo_coeff[0])

        kconserv = np.arange(nkpts)
        moe = scf.addons.mo_energy_with_exxdiv_none(mf)
        e_ia_a = _get_e_ia(moe[0], mo_occ[0], kconserv)
        e_ia_b = _get_e_ia(moe[1], mo_occ[1], kconserv)
        hdiag = np.hstack([x.ravel() for x in (e_ia_a + e_ia_b)])
        tot_x = hdiag.size
        hdiag = np.hstack((hdiag, -hdiag))

        mem_now = lib.current_memory()[0]
        max_memory = max(2000, self.max_memory*.8-mem_now)
        vresp = self.gen_response(hermi=0, max_memory=max_memory)

        def vind(xys):
            nz = len(xys)
            x1s = [_unpack(x[:tot_x], mo_occ, kconserv) for x in xys]
            y1s = [_unpack(x[tot_x:], mo_occ, kconserv) for x in xys]
            dms = np.empty((2,nz,nkpts,nao,nao), dtype=dtype)
            for i in range(nz):
                xa, xb = x1s[i]
                ya, yb = y1s[i]
                for k in range(nkpts):
                    dms[0,i,k]  = lib.einsum('ov,pv,qo->pq', xa[k], orbva[k], orboa[k].conj())
                    dms[1,i,k]  = lib.einsum('ov,pv,qo->pq', xb[k], orbvb[k], orbob[k].conj())
                    dms[0,i,k] += lib.einsum('ov,qv,po->pq', ya[k], orbva[k].conj(), orboa[k])
                    dms[1,i,k] += lib.einsum('ov,qv,po->pq', yb[k], orbvb[k].conj(), orbob[k])

            with lib.temporary_env(mf, exxdiv=None):
                v1ao = vresp(dms, kshift)
                v1ao = v1ao.reshape(2,nz,nkpts,nao,nao)

            v1s = []
            for i in range(nz):
                xa, xb = x1s[i]
                ya, yb = y1s[i]
                v1xsa = [0] * nkpts
                v1xsb = [0] * nkpts
                v1ysa = [0] * nkpts
                v1ysb = [0] * nkpts
                for k in range(nkpts):
                    v1xa = lib.einsum('pq,qo,pv->ov', v1ao[0,i,k], orboa[k], orbva[k].conj())
                    v1xb = lib.einsum('pq,qo,pv->ov', v1ao[1,i,k], orbob[k], orbvb[k].conj())
                    v1ya = lib.einsum('pq,po,qv->ov', v1ao[0,i,k], orboa[k].conj(), orbva[k])
                    v1yb = lib.einsum('pq,po,qv->ov', v1ao[1,i,k], orbob[k].conj(), orbvb[k])
                    v1xa += e_ia_a[k] * xa[k]
                    v1xb += e_ia_b[k] * xb[k]
                    v1ya += e_ia_a[k] * ya[k]
                    v1yb += e_ia_b[k] * yb[k]
                    v1xsa[k] += v1xa.ravel()
                    v1xsb[k] += v1xb.ravel()
                    v1ysa[k] -= v1ya.ravel()
                    v1ysb[k] -= v1yb.ravel()
                v1s.append( np.concatenate(v1xsa + v1xsb + v1ysa + v1ysb) )
            return np.stack(v1s)

        return vind, hdiag

    def get_init_guess(self, mf, kshift, nstates=None, wfnsym=None):
        x0 = TDA.get_init_guess(self, mf, kshift, nstates)
        y0 = np.zeros_like(x0)
        return np.hstack([x0, y0])

    get_precond = uhf.TDHF.get_precond

    def kernel(self, x0=None):
        '''TDHF diagonalization with non-Hermitian eigenvalue solver
        '''
        cpu0 = (logger.process_clock(), logger.perf_counter())
        self.check_sanity()
        self.dump_flags()

        log = logger.new_logger(self)

        mf = self._scf
        mo_occ = mf.mo_occ

        real_system = (is_gamma_point(self._scf.kpts) and
                       self._scf.mo_coeff[0][0].dtype == np.double)

        if any(k != 0 for k in self.kshift_lst):
            raise RuntimeError('kshift != 0 for TDHF')

        # We only need positive eigenvalues
        def pickeig(w, v, nroots, envs):
            realidx = np.where((abs(w.imag) < REAL_EIG_THRESHOLD) &
                                  (w.real > self.positive_eig_threshold))[0]
            return lib.linalg_helper._eigs_cmplx2real(w, v, realidx, real_system)

        self.converged = []
        self.e = []
        self.xy = []
        for i,kshift in enumerate(self.kshift_lst):
            kconserv = get_kconserv_ria(mf.cell, mf.kpts)[kshift]

            vind, hdiag = self.gen_vind(self._scf, kshift)
            precond = self.get_precond(hdiag)

            if x0 is None:
                x0k = self.get_init_guess(self._scf, kshift, self.nstates)
            else:
                x0k = x0[i]

            converged, w, x1 = lr_eig(
                vind, x0k, precond, tol_residual=self.conv_tol, lindep=self.lindep,
                nroots=self.nstates, pick=pickeig, max_cycle=self.max_cycle,
                max_memory=self.max_memory, verbose=log)
            self.converged.append( converged )

            e = []
            xy = []
            for i, z in enumerate(x1):
                xs, ys = z.reshape(2,-1)
                norm = lib.norm(xs)**2 - lib.norm(ys)**2
                if norm < 0:
                    log.warn('TDDFT amplitudes |X| smaller than |Y|')
                norm = abs(norm)**-.5
                xs *= norm
                ys *= norm
                e.append(w[i])
                xy.append((_unpack(xs, mo_occ, kconserv), _unpack(ys, mo_occ, kconserv)))
            self.e.append( np.array(e) )
            self.xy.append( xy )

        log.timer(self.__class__.__name__, *cpu0)
        self._finalize()
        return self.e, self.xy
RPA = KTDHF = TDHF

def _unpack(vo, mo_occ, kconserv):
    za = []
    zb = []
    p1 = 0
    no_a_kpts = [np.count_nonzero(occ) for occ in mo_occ[0]]
    no_b_kpts = [np.count_nonzero(occ) for occ in mo_occ[1]]
    for k, occ in enumerate(mo_occ[0]):
        no = no_a_kpts[k]
        nv = occ.size - no_a_kpts[kconserv[k]]
        p0, p1 = p1, p1 + no * nv
        za.append(vo[p0:p1].reshape(no,nv))

    for k, occ in enumerate(mo_occ[1]):
        no = no_b_kpts[k]
        nv = occ.size - no_b_kpts[kconserv[k]]
        p0, p1 = p1, p1 + no * nv
        zb.append(vo[p0:p1].reshape(no,nv))
    return za, zb


scf.kuhf.KUHF.TDA  = lib.class_as_method(KTDA)
scf.kuhf.KUHF.TDHF = lib.class_as_method(KTDHF)



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''Density expansion on plane waves'''

import numpy
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.pbc import tools
from pyscf.pbc import gto as pbcgto
from pyscf.pbc.gto import pseudo, error_for_ke_cutoff, estimate_ke_cutoff
from pyscf.pbc.df import ft_ao
from pyscf.pbc.df import fft_ao2mo
from pyscf.pbc.df import fft_jk
from pyscf.pbc.df import aft
from pyscf.pbc.df.aft import _check_kpts
from pyscf.pbc.lib.kpts import KPoints
from pyscf.pbc.lib.kpts_helper import is_zero
from pyscf import __config__

KE_SCALING = getattr(__config__, 'pbc_df_aft_ke_cutoff_scaling', 0.75)


def get_nuc(mydf, kpts=None):
    from pyscf.pbc.dft import gen_grid
    kpts, is_single_kpt = _check_kpts(mydf, kpts)
    cell = mydf.cell
    assert cell.low_dim_ft_type != 'inf_vacuum'
    assert cell.dimension != 1
    mesh = mydf.mesh
    charge = -cell.atom_charges()
    Gv = cell.get_Gv(mesh)
    SI = cell.get_SI(mesh=mesh)
    rhoG = numpy.dot(charge, SI)

    coulG = tools.get_coulG(cell, mesh=mesh, Gv=Gv)
    vneG = rhoG * coulG
    vneR = tools.ifft(vneG, mesh).real

    vne = [0] * len(kpts)
    for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts):
        ao_ks = ao_ks_etc[0]
        for k, ao in enumerate(ao_ks):
            vne[k] += lib.dot(ao.T.conj()*vneR[p0:p1], ao)
        ao = ao_ks = None

    if is_single_kpt:
        vne = vne[0]
    return numpy.asarray(vne)

def get_pp(mydf, kpts=None):
    '''Get the periodic pseudopotential nuc-el AO matrix, with G=0 removed.
    '''
    from pyscf.pbc.dft import gen_grid
    kpts, is_single_kpt = _check_kpts(mydf, kpts)
    cell = mydf.cell
    assert cell.low_dim_ft_type != 'inf_vacuum'
    assert cell.dimension != 1
    mesh = mydf.mesh
    Gv = cell.get_Gv(mesh)
    SI = cell.get_SI(mesh=mesh)
    vpplocG = pseudo.get_vlocG(cell, Gv)
    vpplocG = -numpy.einsum('ij,ij->j', SI, vpplocG)
    ngrids = len(vpplocG)

    # vpploc evaluated in real-space
    vpplocR = tools.ifft(vpplocG, mesh).real
    vpp = [0] * len(kpts)
    for ao_ks_etc, p0, p1 in mydf.aoR_loop(mydf.grids, kpts):
        ao_ks = ao_ks_etc[0]
        for k, ao in enumerate(ao_ks):
            vpp[k] += lib.dot(ao.T.conj()*vpplocR[p0:p1], ao)
        ao = ao_ks = None

    # vppnonloc evaluated in reciprocal space
    fakemol = gto.Mole()
    fakemol._atm = numpy.zeros((1,gto.ATM_SLOTS), dtype=numpy.int32)
    fakemol._bas = numpy.zeros((1,gto.BAS_SLOTS), dtype=numpy.int32)
    ptr = gto.PTR_ENV_START
    fakemol._env = numpy.zeros(ptr+10)
    fakemol._bas[0,gto.NPRIM_OF ] = 1
    fakemol._bas[0,gto.NCTR_OF  ] = 1
    fakemol._bas[0,gto.PTR_EXP  ] = ptr+3
    fakemol._bas[0,gto.PTR_COEFF] = ptr+4

    # buf for SPG_lmi upto l=0..3 and nl=3
    buf = numpy.empty((48,ngrids), dtype=numpy.complex128)
    def vppnl_by_k(kpt):
        Gk = Gv + kpt
        G_rad = lib.norm(Gk, axis=1)
        aokG = ft_ao.ft_ao(cell, Gv, kpt=kpt) * (1/cell.vol)**.5
        vppnl = 0
        for ia in range(cell.natm):
            symb = cell.atom_symbol(ia)
            if symb not in cell._pseudo:
                continue
            pp = cell._pseudo[symb]
            p1 = 0
            for l, proj in enumerate(pp[5:]):
                rl, nl, hl = proj
                if nl > 0:
                    fakemol._bas[0,gto.ANG_OF] = l
                    fakemol._env[ptr+3] = .5*rl**2
                    fakemol._env[ptr+4] = rl**(l+1.5)*numpy.pi**1.25
                    pYlm_part = fakemol.eval_gto('GTOval', Gk)

                    p0, p1 = p1, p1+nl*(l*2+1)
                    # pYlm is real, SI[ia] is complex
                    pYlm = numpy.ndarray((nl,l*2+1,ngrids), dtype=numpy.complex128, buffer=buf[p0:p1])
                    for k in range(nl):
                        qkl = pseudo.pp._qli(G_rad*rl, l, k)
                        pYlm[k] = pYlm_part.T * qkl
                    #:SPG_lmi = numpy.einsum('g,nmg->nmg', SI[ia].conj(), pYlm)
                    #:SPG_lm_aoG = numpy.einsum('nmg,gp->nmp', SPG_lmi, aokG)
                    #:tmp = numpy.einsum('ij,jmp->imp', hl, SPG_lm_aoG)
                    #:vppnl += numpy.einsum('imp,imq->pq', SPG_lm_aoG.conj(), tmp)
            if p1 > 0:
                SPG_lmi = buf[:p1]
                SPG_lmi *= SI[ia].conj()
                SPG_lm_aoGs = lib.zdot(SPG_lmi, aokG)
                p1 = 0
                for l, proj in enumerate(pp[5:]):
                    rl, nl, hl = proj
                    if nl > 0:
                        p0, p1 = p1, p1+nl*(l*2+1)
                        hl = numpy.asarray(hl)
                        SPG_lm_aoG = SPG_lm_aoGs[p0:p1].reshape(nl,l*2+1,-1)
                        tmp = numpy.einsum('ij,jmp->imp', hl, SPG_lm_aoG)
                        vppnl += numpy.einsum('imp,imq->pq', SPG_lm_aoG.conj(), tmp)
        return vppnl * (1./cell.vol)

    for k, kpt in enumerate(kpts):
        vppnl = vppnl_by_k(kpt)
        if is_zero(kpt):
            vpp[k] = vpp[k].real + vppnl.real
        else:
            vpp[k] += vppnl

    if is_single_kpt:
        vpp = vpp[0]
    return numpy.asarray(vpp)


class FFTDF(lib.StreamObject):
    '''Density expansion on plane waves
    '''

    # to mimic molecular DF object
    blockdim = getattr(__config__, 'pbc_df_df_DF_blockdim', 240)

    _keys = {
        'cell', 'kpts', 'grids', 'mesh', 'blockdim', 'exxdiv',
    }

    def __init__(self, cell, kpts=numpy.zeros((1,3))):
        from pyscf.pbc.dft import numint
        self.cell = cell
        self.stdout = cell.stdout
        self.verbose = cell.verbose
        self.max_memory = cell.max_memory

        if isinstance(kpts, KPoints):
            kpts = kpts.kpts
        self.kpts = kpts
        # FFT from real space density distributes error to every rho_ij(G) than
        # the one with largest Gaussian exponent. Therefore the error for FFT-ERI
        # ~ Nele * error[rho(Ecut)] while in AFT the error is ~ error[rho(Ecut)]^2.
        # This is a first order error, same to the error estimation for nuclear
        # attraction.
        self.mesh = cell.mesh

        # The following attributes are not input options.
        # self.exxdiv has no effects. It was set in the get_k_kpts function to
        # mimic the KRHF/KUHF object in the call to tools.get_coulG.
        self.exxdiv = None
        self._numint = numint.KNumInt()
        self._rsh_df = {}  # Range separated Coulomb DF objects

    @property
    def grids(self):
        from pyscf.pbc.dft import gen_grid
        grids = gen_grid.UniformGrids(self.cell)
        grids.mesh = self.mesh
        return grids

    def reset(self, cell=None):
        if cell is not None:
            self.cell = cell
        self._rsh_df = {}
        return self

    def dump_flags(self, verbose=None):
        logger.info(self, '\n')
        logger.info(self, '******** %s ********', self.__class__)
        logger.info(self, 'mesh = %s (%d PWs)', self.mesh, numpy.prod(self.mesh))
        logger.info(self, 'len(kpts) = %d', len(self.kpts))
        logger.debug1(self, '    kpts = %s', self.kpts)
        return self

    def check_sanity(self):
        lib.StreamObject.check_sanity(self)
        cell = self.cell
        if (cell.dimension < 2 or
            (cell.dimension == 2 and cell.low_dim_ft_type == 'inf_vacuum')):
            raise RuntimeError('FFTDF method does not support 0D/1D low-dimension '
                               'PBC system.  DF, MDF or AFTDF methods should '
                               'be used.\nSee also examples/pbc/31-low_dimensional_pbc.py')
        if not cell.has_ecp():
            logger.warn(self, 'FFTDF integrals are found in all-electron '
                        'calculation.  It often causes huge error.\n'
                        'Recommended methods are DF or MDF. In SCF calculation, '
                        'they can be initialized as\n'
                        '        mf = mf.density_fit()\nor\n'
                        '        mf = mf.mix_density_fit()')

        if cell.ke_cutoff is None:
            ke_cutoff = tools.mesh_to_cutoff(cell.lattice_vectors(), self.mesh).min()
        else:
            ke_cutoff = numpy.min(cell.ke_cutoff)
        ke_guess = estimate_ke_cutoff(cell, cell.precision)
        if ke_cutoff < ke_guess * KE_SCALING:
            mesh_guess = cell.cutoff_to_mesh(ke_guess)
            logger.warn(self, 'ke_cutoff/mesh (%g / %s) is not enough for FFTDF '
                        'to get integral accuracy %g.\nCoulomb integral error '
                        'is ~ %.2g Eh.\nRecommended ke_cutoff/mesh are %g / %s.',
                        ke_cutoff, self.mesh, cell.precision,
                        error_for_ke_cutoff(cell, ke_cutoff), ke_guess, mesh_guess)
        return self

    def build(self):
        return self.check_sanity()

    def aoR_loop(self, grids=None, kpts=None, deriv=0):
        if grids is None:
            grids = self.grids
            cell = self.cell
        else:
            cell = grids.cell
        if grids.non0tab is None:
            grids.build(with_non0tab=True)

        if kpts is None: kpts = self.kpts
        kpts = numpy.asarray(kpts)

        if cell.dimension <= 2 and cell.low_dim_ft_type == 'inf_vacuum':
            raise RuntimeError('FFTDF method does not support low-dimension '
                               'PBC system.  DF, MDF or AFTDF methods should '
                               'be used.\nSee also examples/pbc/31-low_dimensional_pbc.py')

        max_memory = max(2000, self.max_memory-lib.current_memory()[0])
        ni = self._numint
        nao = cell.nao_nr()
        p1 = 0
        for ao_k1_etc in ni.block_loop(cell, grids, nao, deriv, kpts,
                                       max_memory=max_memory):
            coords = ao_k1_etc[4]
            p0, p1 = p1, p1 + coords.shape[0]
            yield ao_k1_etc, p0, p1

    get_pp = get_pp
    get_nuc = get_nuc

    def get_jk_e1(self, dm, kpts=None, kpts_band=None, exxdiv=None):
        kpts = _check_kpts(self, kpts)[0]
        vj = fft_jk.get_j_e1_kpts(self, dm, kpts, kpts_band)
        vk = fft_jk.get_k_e1_kpts(self, dm, kpts, kpts_band, exxdiv)
        return vj, vk

    def get_j_e1(self, dm, kpts=None, kpts_band=None):
        kpts = _check_kpts(self, kpts)[0]
        vj = fft_jk.get_j_e1_kpts(self, dm, kpts, kpts_band)
        return vj

    def get_k_e1(self, dm, kpts=None, kpts_band=None, exxdiv=None):
        kpts = _check_kpts(self, kpts)[0]
        vk = fft_jk.get_k_e1_kpts(self, dm, kpts, kpts_band, exxdiv)
        return vk

    # Note: Special exxdiv by default should not be used for an arbitrary
    # input density matrix. When the df object was used with the molecular
    # post-HF code, get_jk was often called with an incomplete DM (e.g. the
    # core DM in CASCI). An SCF level exxdiv treatment is inadequate for
    # post-HF methods.
    def get_jk(self, dm, hermi=1, kpts=None, kpts_band=None,
               with_j=True, with_k=True, omega=None, exxdiv=None):
        if omega is not None:  # J/K for RSH functionals
            with self.range_coulomb(omega) as rsh_df:
                return rsh_df.get_jk(dm, hermi, kpts, kpts_band, with_j, with_k,
                                     omega=None, exxdiv=exxdiv)

        kpts, is_single_kpt = _check_kpts(self, kpts)
        if is_single_kpt:
            vj, vk = fft_jk.get_jk(self, dm, hermi, kpts[0], kpts_band,
                                   with_j, with_k, exxdiv)
        else:
            vj = vk = None
            if with_k:
                vk = fft_jk.get_k_kpts(self, dm, hermi, kpts, kpts_band, exxdiv)
            if with_j:
                vj = fft_jk.get_j_kpts(self, dm, hermi, kpts, kpts_band)
        return vj, vk

    get_eri = get_ao_eri = fft_ao2mo.get_eri
    ao2mo = get_mo_eri = fft_ao2mo.general
    ao2mo_7d = fft_ao2mo.ao2mo_7d
    get_ao_pairs_G = get_ao_pairs = fft_ao2mo.get_ao_pairs_G
    get_mo_pairs_G = get_mo_pairs = fft_ao2mo.get_mo_pairs_G

    def update_mf(self, mf):
        mf = mf.copy()
        mf.with_df = self
        return mf

################################################################################
# With this function to mimic the molecular DF.loop function, the pbc gamma
# point DF object can be used in the molecular code
    def loop(self, blksize=None):
        if self.cell.dimension < 3:
            raise RuntimeError('ERIs of 1D and 2D systems are not positive '
                               'definite. Current API only supports positive '
                               'definite ERIs.')

        if blksize is None:
            blksize = self.blockdim
        kpts0 = numpy.zeros((2,3))
        coulG = tools.get_coulG(self.cell, numpy.zeros(3), mesh=self.mesh)
        ngrids = len(coulG)
        ao_pairs_G = self.get_ao_pairs_G(kpts0, compact=True)
        ao_pairs_G *= numpy.sqrt(coulG*(self.cell.vol/ngrids**2)).reshape(-1,1)

        Lpq = numpy.empty((blksize, ao_pairs_G.shape[1]))
        for p0, p1 in lib.prange(0, ngrids, blksize):
            Lpq[:p1-p0] = ao_pairs_G[p0:p1].real
            yield Lpq[:p1-p0]
            Lpq[:p1-p0] = ao_pairs_G[p0:p1].imag
            yield Lpq[:p1-p0]

    def get_naoaux(self):
        mesh = numpy.asarray(self.mesh)
        ngrids = numpy.prod(mesh)
        return ngrids * 2

    range_coulomb = aft.AFTDF.range_coulomb

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#         Yu Jin <jinyuchem@uchicago.edu>
#

import ctypes
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.cc import ccsd_rdm, _ccsd

def _gamma1_intermediates(mycc, t1, t2, l1, l2, eris=None, for_grad=False):
    log = logger.Logger(mycc.stdout, mycc.verbose)

    if (numpy.iscomplexobj(t1) or numpy.iscomplexobj(t2) or numpy.iscomplexobj(eris)
        or numpy.iscomplexobj(l1) or numpy.iscomplexobj(l2)):
        raise ValueError("_gamma1_intermediates does not support complex-valued inputs (t1, t2, l1, l2, or eris)")

    doo, dov, dvo, dvv = ccsd_rdm._gamma1_intermediates(mycc, t1, t2, l1, l2)

    if eris is None: eris = mycc.ao2mo()
    nocc, nvir = t1.shape
    eris_ovvv = numpy.asarray(eris.get_ovvv())
    eris_ovoo = numpy.asarray(eris.ovoo)
    eris_ovov = numpy.asarray(eris.ovov)

    mo_e = eris.mo_energy
    eia = lib.direct_sum('i-a->ia', mo_e[:nocc], mo_e[nocc:])

    mem_now = lib.current_memory()[0]
    max_memory = max(0, mycc.max_memory - mem_now)
    blksize = min(nvir, int(((max_memory * 0.9e6/8) / 6.0 / (nocc**3))**(1/3)))
    if blksize < nvir:
        blksize = min(blksize, (nvir + 1) // 2)
        blksize = max(blksize, 1)
    log.debug1('ccsd_t rdm _gamma1_intermediates: max_memory %d MB,  nocc,nvir = %d,%d  blksize = %d',
               max_memory, nocc, nvir, blksize)

    goo = numpy.zeros((nocc, nocc), dtype=t1.dtype)
    w_blk = numpy.empty((blksize, blksize, blksize, nocc, nocc, nocc), dtype=t1.dtype)
    v_blk = numpy.empty((blksize, blksize, blksize, nocc, nocc, nocc), dtype=t1.dtype)

    time2 = logger.process_clock(), logger.perf_counter()
    for a0, a1 in lib.prange(0, nvir, blksize):
        ba = a1 - a0
        for b0, b1 in lib.prange(0, nvir, blksize):
            bb = b1 - b0
            for c0, c1 in lib.prange(0, nvir, blksize):
                bc = c1 - c0

                w_blk[:ba, :bb, :bc] = lib.einsum('iafb,kjcf->abcijk',
                    eris_ovvv[:, a0:a1, :, b0:b1], t2[:, :, c0:c1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('iafc,jkbf->abcijk',
                    eris_ovvv[:, a0:a1, :, c0:c1], t2[:, :, b0:b1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('jbfa,kicf->abcijk',
                    eris_ovvv[:, b0:b1, :, a0:a1], t2[:, :, c0:c1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('jbfc,ikaf->abcijk',
                    eris_ovvv[:, b0:b1, :, c0:c1], t2[:, :, a0:a1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('kcfa,jibf->abcijk',
                    eris_ovvv[:, c0:c1, :, a0:a1], t2[:, :, b0:b1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('kcfb,ijaf->abcijk',
                    eris_ovvv[:, c0:c1, :, b0:b1], t2[:, :, a0:a1, :])
                w_blk[:ba, :bb, :bc] -= lib.einsum('iajm,mkbc->abcijk',
                    eris_ovoo[:, a0:a1, :, :], t2[:, :, b0:b1, c0:c1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('iakm,mjcb->abcijk',
                    eris_ovoo[:, a0:a1, :, :], t2[:, :, c0:c1, b0:b1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('jbim,mkac->abcijk',
                    eris_ovoo[:, b0:b1, :, :], t2[:, :, a0:a1, c0:c1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('jbkm,mica->abcijk',
                    eris_ovoo[:, b0:b1, :, :], t2[:, :, c0:c1, a0:a1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('kcim,mjab->abcijk',
                    eris_ovoo[:, c0:c1, :, :], t2[:, :, a0:a1, b0:b1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('kcjm,miba->abcijk',
                    eris_ovoo[:, c0:c1, :, :], t2[:, :, b0:b1, a0:a1])

                v_blk[:ba, :bb, :bc] = lib.einsum('iajb,kc->abcijk',
                    eris_ovov[:, a0:a1, :, b0:b1], t1[:, c0:c1])
                v_blk[:ba, :bb, :bc] += lib.einsum('iakc,jb->abcijk',
                    eris_ovov[:, a0:a1, :, c0:c1], t1[:, b0:b1])
                v_blk[:ba, :bb, :bc] += lib.einsum('jbkc,ia->abcijk',
                    eris_ovov[:, b0:b1, :, c0:c1], t1[:, a0:a1])
                v_blk[:ba, :bb, :bc] += lib.einsum('ck,ijab->abcijk',
                    eris.fock[(nocc + c0):(nocc + c1), :nocc], t2[:, :, a0:a1, b0:b1])
                v_blk[:ba, :bb, :bc] += lib.einsum('ai,jkbc->abcijk',
                    eris.fock[(nocc + a0):(nocc + a1), :nocc], t2[:, :, b0:b1, c0:c1])
                v_blk[:ba, :bb, :bc] += lib.einsum('bj,kica->abcijk',
                    eris.fock[(nocc + b0):(nocc + b1), :nocc], t2[:, :, c0:c1, a0:a1])

                d3_blk = lib.direct_sum('ia,jb,kc->abcijk', eia[:, a0:a1], eia[:, b0:b1], eia[:, c0:c1])
                w_blk[:ba, :bb, :bc] /= d3_blk
                v_blk[:ba, :bb, :bc] /= d3_blk

                v_blk += w_blk

                t3_symm_ip_py(w_blk, blksize**3, nocc, "4-2-211-2", 1.0, 0.0)
                goo[:, :] += lib.einsum('abcikl,abcjkl->ij',
                    v_blk[:ba, :bb, :bc, :, :, :], w_blk[:ba, :bb, :bc, :, :, :])

        time2 = log.timer_debug1('ccsd_t rdm _gamma1_intermediates pass1 [%d:%d]'%(a0, a1), *time2)

    w_blk = None
    v_blk = None

    mem_now = lib.current_memory()[0]
    max_memory = max(0, mycc.max_memory - mem_now)
    blksize = min(nocc, int(((max_memory * 0.9e6/8) / 6.0 / (nvir**3))**(1/3)))
    if blksize < nocc:
        blksize = min(blksize, (nocc + 1) // 2)
        blksize = max(blksize, 1)
    log.debug1('ccsd_t rdm _gamma1_intermediates: max_memory %d MB,  nocc,nvir = %d,%d  blksize = %d',
               max_memory, nocc, nvir, blksize)

    gvv = numpy.zeros((nvir, nvir), dtype=t1.dtype)
    w_blk = numpy.empty((blksize, blksize, blksize, nvir, nvir, nvir), dtype=t1.dtype)
    v_blk = numpy.empty((blksize, blksize, blksize, nvir, nvir, nvir), dtype=t1.dtype)

    time2 = logger.process_clock(), logger.perf_counter()
    for k0, k1 in lib.prange(0, nocc, blksize):
        bk = k1 - k0
        for j0, j1 in lib.prange(0, nocc, blksize):
            bj = j1 - j0
            for i0, i1 in lib.prange(0, nocc, blksize):
                bi = i1 - i0

                w_blk[:bi, :bj, :bk] = lib.einsum('iafb,kjcf->ijkabc',
                    eris_ovvv[i0:i1, :, :, :], t2[k0:k1, j0:j1, :, :])
                w_blk[:bi, :bj, :bk] += lib.einsum('iafc,jkbf->ijkabc',
                    eris_ovvv[i0:i1, :, :, :], t2[j0:j1, k0:k1, :, :])
                w_blk[:bi, :bj, :bk] += lib.einsum('jbfa,kicf->ijkabc',
                    eris_ovvv[j0:j1, :, :, :], t2[k0:k1, i0:i1, :, :])
                w_blk[:bi, :bj, :bk] += lib.einsum('jbfc,ikaf->ijkabc',
                    eris_ovvv[j0:j1, :, :, :], t2[i0:i1, k0:k1, :, :])
                w_blk[:bi, :bj, :bk] += lib.einsum('kcfa,jibf->ijkabc',
                    eris_ovvv[k0:k1, :, :, :], t2[j0:j1, i0:i1, :, :])
                w_blk[:bi, :bj, :bk] += lib.einsum('kcfb,ijaf->ijkabc',
                    eris_ovvv[k0:k1, :, :, :], t2[i0:i1, j0:j1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('iajm,mkbc->ijkabc',
                    eris_ovoo[i0:i1, :, j0:j1, :], t2[:, k0:k1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('iakm,mjcb->ijkabc',
                    eris_ovoo[i0:i1, :, k0:k1, :], t2[:, j0:j1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('jbim,mkac->ijkabc',
                    eris_ovoo[j0:j1, :, i0:i1, :], t2[:, k0:k1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('jbkm,mica->ijkabc',
                    eris_ovoo[j0:j1, :, k0:k1, :], t2[:, i0:i1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('kcim,mjab->ijkabc',
                    eris_ovoo[k0:k1, :, i0:i1, :], t2[:, j0:j1, :, :])
                w_blk[:bi, :bj, :bk] -= lib.einsum('kcjm,miba->ijkabc',
                    eris_ovoo[k0:k1, :, j0:j1, :], t2[:, i0:i1, :, :])

                v_blk[:bi, :bj, :bk] = lib.einsum('iajb,kc->ijkabc',
                    eris_ovov[i0:i1, :, j0:j1, :], t1[k0:k1, :])
                v_blk[:bi, :bj, :bk] += lib.einsum('iakc,jb->ijkabc',
                    eris_ovov[i0:i1, :, k0:k1, :], t1[j0:j1, :])
                v_blk[:bi, :bj, :bk] += lib.einsum('jbkc,ia->ijkabc',
                    eris_ovov[j0:j1, :, k0:k1, :], t1[i0:i1, :])
                v_blk[:bi, :bj, :bk] += lib.einsum('ck,ijab->ijkabc',
                    eris.fock[nocc:, k0:k1], t2[i0:i1, j0:j1, :, :])
                v_blk[:bi, :bj, :bk] += lib.einsum('ai,jkbc->ijkabc',
                    eris.fock[nocc:, i0:i1], t2[j0:j1, k0:k1, :, :])
                v_blk[:bi, :bj, :bk] += lib.einsum('bj,kica->ijkabc',
                    eris.fock[nocc:, j0:j1], t2[k0:k1, i0:i1, :, :])

                d3_blk = lib.direct_sum('ia,jb,kc->ijkabc', eia[i0:i1, :], eia[j0:j1, :], eia[k0:k1, :])
                w_blk[:bi, :bj, :bk] /= d3_blk
                v_blk[:bi, :bj, :bk] /= d3_blk

                v_blk += w_blk

                t3_symm_ip_py(w_blk, blksize**3, nvir, "4-2-211-2", 1.0, 0.0)

                gvv[:, :] += lib.einsum('ijkacd,ijkbcd->ab',
                    v_blk[:bi, :bj, :bk, :, :, :], w_blk[:bi, :bj, :bk, :, :, :])
                dvo[:, k0:k1] += 0.5 * lib.einsum('ijab,ijkabc->ck',
                    t2[i0:i1, j0:j1, :, :], w_blk[:bi, :bj, :bk, :, :, :])

        time2 = log.timer_debug1('ccsd_t rdm _gamma1_intermediates pass2 [%d:%d]'%(k0, k1), *time2)

    w_blk = None
    v_blk = None

    if not for_grad:
        # t3 amplitudes in CCSD(T) is computed non-iteratively. The
        # off-diagonal blocks of fock matrix does not contribute to CCSD(T)
        # energy. To make Tr(H,D) consistent to the CCSD(T) total energy, the
        # density matrix off-diagonal parts are excluded.
        doo[numpy.diag_indices(nocc)] -= goo.diagonal() * .5
        dvv[numpy.diag_indices(nvir)] += gvv.diagonal() * .5

    else:
        # The off-diagonal blocks of fock matrix have small contributions to
        # analytical nuclear gradients.
        doo -= goo * .5
        dvv += gvv * .5

    return doo, dov, dvo, dvv

def _gamma2_intermediates(mycc, t1, t2, l1, l2, eris=None,
                          compress_vvvv=False):
    '''intermediates tensors for gamma2 are sorted in Chemist's notation
    '''
    log = logger.Logger(mycc.stdout, mycc.verbose)

    if (numpy.iscomplexobj(t1) or numpy.iscomplexobj(t2) or numpy.iscomplexobj(eris)
        or numpy.iscomplexobj(l1) or numpy.iscomplexobj(l2)):
        raise ValueError("_gamma2_intermediates does not support complex-valued inputs (t1, t2, l1, l2, or eris)")

    dovov, dvvvv, doooo, doovv, dovvo, dvvov, dovvv, dooov = \
            ccsd_rdm._gamma2_intermediates(mycc, t1, t2, l1, l2)
    if eris is None: eris = mycc.ao2mo()

    nocc, nvir = t1.shape
    eris_ovvv = numpy.asarray(eris.get_ovvv())
    eris_ovoo = numpy.asarray(eris.ovoo)
    eris_ovov = numpy.asarray(eris.ovov)

    mo_e = eris.mo_energy
    eia = lib.direct_sum('i-a->ia', mo_e[:nocc], mo_e[nocc:])

    mem_now = lib.current_memory()[0]
    max_memory = max(0, mycc.max_memory - mem_now)
    blksize = min(nvir, int(((max_memory * 0.9e6/8) / 6.0 / (nocc**3))**(1/3)))
    if blksize < nvir:
        blksize = min(blksize, (nvir + 1) // 2)
        blksize = max(blksize, 1)
    log.debug1('ccsd_t rdm _gamma2_intermediates: max_memory %d MB,  nocc,nvir = %d,%d  blksize = %d',
               max_memory, nocc, nvir, blksize)

    w_blk = numpy.empty((blksize, blksize, blksize, nocc, nocc, nocc), dtype=t1.dtype)
    v_blk = numpy.empty((blksize, blksize, blksize, nocc, nocc, nocc), dtype=t1.dtype)

    time2 = logger.process_clock(), logger.perf_counter()
    for c0, c1 in lib.prange(0, nvir, blksize):
        bc = c1 - c0
        for b0, b1 in lib.prange(0, nvir, blksize):
            bb = b1 - b0
            for a0, a1 in lib.prange(0, nvir, blksize):
                ba = a1 - a0

                w_blk[:ba, :bb, :bc] = lib.einsum('iafb,kjcf->abcijk',
                    eris_ovvv[:, a0:a1, :, b0:b1], t2[:, :, c0:c1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('iafc,jkbf->abcijk',
                    eris_ovvv[:, a0:a1, :, c0:c1], t2[:, :, b0:b1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('jbfa,kicf->abcijk',
                    eris_ovvv[:, b0:b1, :, a0:a1], t2[:, :, c0:c1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('jbfc,ikaf->abcijk',
                    eris_ovvv[:, b0:b1, :, c0:c1], t2[:, :, a0:a1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('kcfa,jibf->abcijk',
                    eris_ovvv[:, c0:c1, :, a0:a1], t2[:, :, b0:b1, :])
                w_blk[:ba, :bb, :bc] += lib.einsum('kcfb,ijaf->abcijk',
                    eris_ovvv[:, c0:c1, :, b0:b1], t2[:, :, a0:a1, :])
                w_blk[:ba, :bb, :bc] -= lib.einsum('iajm,mkbc->abcijk',
                    eris_ovoo[:, a0:a1, :, :], t2[:, :, b0:b1, c0:c1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('iakm,mjcb->abcijk',
                    eris_ovoo[:, a0:a1, :, :], t2[:, :, c0:c1, b0:b1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('jbim,mkac->abcijk',
                    eris_ovoo[:, b0:b1, :, :], t2[:, :, a0:a1, c0:c1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('jbkm,mica->abcijk',
                    eris_ovoo[:, b0:b1, :, :], t2[:, :, c0:c1, a0:a1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('kcim,mjab->abcijk',
                    eris_ovoo[:, c0:c1, :, :], t2[:, :, a0:a1, b0:b1])
                w_blk[:ba, :bb, :bc] -= lib.einsum('kcjm,miba->abcijk',
                    eris_ovoo[:, c0:c1, :, :], t2[:, :, b0:b1, a0:a1])

                v_blk[:ba, :bb, :bc] = lib.einsum('iajb,kc->abcijk',
                    eris_ovov[:, a0:a1, :, b0:b1], t1[:, c0:c1])
                v_blk[:ba, :bb, :bc] += lib.einsum('iakc,jb->abcijk',
                    eris_ovov[:, a0:a1, :, c0:c1], t1[:, b0:b1])
                v_blk[:ba, :bb, :bc] += lib.einsum('jbkc,ia->abcijk',
                    eris_ovov[:, b0:b1, :, c0:c1], t1[:, a0:a1])
                v_blk[:ba, :bb, :bc] += lib.einsum('ck,ijab->abcijk',
                    eris.fock[nocc + c0:nocc + c1, :nocc], t2[:, :, a0:a1, b0:b1])
                v_blk[:ba, :bb, :bc] += lib.einsum('ai,jkbc->abcijk',
                    eris.fock[nocc + a0:nocc + a1, :nocc], t2[:, :, b0:b1, c0:c1])
                v_blk[:ba, :bb, :bc] += lib.einsum('bj,kica->abcijk',
                    eris.fock[nocc + b0:nocc + b1, :nocc], t2[:, :, c0:c1, a0:a1])

                d3_blk = lib.direct_sum('ia,jb,kc->abcijk', eia[:, a0:a1], eia[:, b0:b1], eia[:, c0:c1])
                w_blk[:ba, :bb, :bc] /= d3_blk
                v_blk[:ba, :bb, :bc] /= d3_blk

                v_blk += 2.0 * w_blk
                t3_symm_ip_py(v_blk, blksize**3, nocc, "4-2-211-2", 1.0, 0.0)
                t3_symm_ip_py(w_blk, blksize**3, nocc, "4-2-211-2", 0.5, 0.0)

                dovov[:, a0:a1, :, b0:b1] += lib.einsum('kc,abcijk->iajb',
                    t1[:, c0:c1], w_blk[:ba, :bb, :bc, :, :, :])
                dooov[:, :, :, a0:a1] -= lib.einsum('mkbc,abcijk->jmia',
                    t2[:, :, b0:b1, c0:c1], v_blk[:ba, :bb, :bc, :, :, :])
                dovvv[:, a0:a1, :, b0:b1] += lib.einsum('kjcf,abcijk->iafb',
                    t2[:, :, c0:c1, :], v_blk[:ba, :bb, :bc, :, :, :])

        time2 = log.timer_debug1('ccsd_t rdm _gamma2_intermediates [%d:%d]'%(c0, c1), *time2)

    w_blk = None
    v_blk = None

    dvvov = dovvv.transpose(2,3,0,1)

    if compress_vvvv:
        nvir = mycc.nmo - mycc.nocc
        idx = numpy.tril_indices(nvir)
        vidx = idx[0] * nvir + idx[1]
        dvvvv = dvvvv + dvvvv.transpose(1,0,2,3)
        dvvvv = dvvvv + dvvvv.transpose(0,1,3,2)
        dvvvv = lib.take_2d(dvvvv.reshape(nvir**2,nvir**2), vidx, vidx)
        dvvvv *= .25

    return dovov, dvvvv, doooo, doovv, dovvo, dvvov, dovvv, dooov

def _gamma2_outcore(mycc, t1, t2, l1, l2, eris, h5fobj, compress_vvvv=False):
    return _gamma2_intermediates(mycc, t1, t2, l1, l2, eris, compress_vvvv)

def make_rdm1(mycc, t1, t2, l1, l2, eris=None, ao_repr=False):
    d1 = _gamma1_intermediates(mycc, t1, t2, l1, l2, eris)
    return ccsd_rdm._make_rdm1(mycc, d1, True, ao_repr=ao_repr)

# rdm2 in Chemist's notation
def make_rdm2(mycc, t1, t2, l1, l2, eris=None):
    d1 = _gamma1_intermediates(mycc, t1, t2, l1, l2, eris)
    d2 = _gamma2_intermediates(mycc, t1, t2, l1, l2, eris)
    return ccsd_rdm._make_rdm2(mycc, d1, d2, True, True)

def t3_symm_ip_py(A, nocc3, nvir, pattern, alpha=1.0, beta=0.0):
    assert A.dtype == numpy.float64 and A.flags['C_CONTIGUOUS'], "A must be a contiguous float64 array"

    pattern_c = pattern.encode('utf-8')

    drv = _ccsd.libcc.t3_symm_ip
    drv(
        A.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int64(nocc3),
        ctypes.c_int64(nvir),
        ctypes.c_char_p(pattern_c),
        ctypes.c_double(alpha),
        ctypes.c_double(beta)
    )
    return A



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from pyscf import lib

def _get_jk_kshift(mf, dm_kpts, hermi, kpts, kshift, with_j=True, with_k=True,
                   omega=None):
    from pyscf.pbc.df.df_jk import get_j_kpts_kshift, get_k_kpts_kshift
    vj = vk = None
    if with_j:
        vj = get_j_kpts_kshift(mf.with_df, dm_kpts, kshift, hermi=hermi, kpts=kpts)
    if with_k:
        vk = get_k_kpts_kshift(mf.with_df, dm_kpts, kshift, hermi=hermi, kpts=kpts,
                               exxdiv=mf.exxdiv)
    return vj, vk
def _get_jk(mf, cell, dm1, hermi, kpts, kshift, with_j=True, with_k=True, omega=None):
    from pyscf.pbc import df
    if kshift == 0:
        return mf.get_jk(cell, dm1, hermi=hermi, kpts=kpts,
                         with_j=with_j, with_k=with_k, omega=omega)
    elif omega is not None and omega != 0:
        raise NotImplementedError
    elif mf.rsjk is not None or not isinstance(mf.with_df, df.df.DF):
        lib.logger.error(mf, 'Non-zero kshift is only supported by GDF/RSDF.')
        raise NotImplementedError
    else:
        return _get_jk_kshift(mf, dm1, hermi, kpts, kshift,
                              with_j=with_j, with_k=with_k, omega=omega)
def _get_j(mf, cell, dm1, hermi, kpts, kshift, omega=None):
    return _get_jk(mf, cell, dm1, hermi, kpts, kshift, True, False, omega)[0]
def _get_k(mf, cell, dm1, hermi, kpts, kshift, omega=None):
    return _get_jk(mf, cell, dm1, hermi, kpts, kshift, False, True, omega)[1]



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Analytical Fourier transformation for AO and AO-pair value
'''

import ctypes
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import gto
from pyscf.gto.moleintor import libcgto

#
# \int mu*nu*exp(-ik*r) dr
#
# gxyz is the index for Gvbase
def ft_aopair(mol, Gv, shls_slice=None, aosym='s1', b=numpy.eye(3),
              gxyz=None, Gvbase=None, out=None, intor='GTO_ft_ovlp',
              comp=1, q=numpy.zeros(3), return_complex=True, ovlp_mask=None,
              verbose=None):
    r''' FT transform AO pair
    \int i(r) j(r) exp(-ikr) dr^3
    '''

    intor = mol._add_suffix(intor)

    if shls_slice is None:
        shls_slice = (0, mol.nbas, 0, mol.nbas)

    GvT = numpy.asarray(Gv.T, order='C') + q[:,None]
    nGv = Gv.shape[0]
    if (gxyz is None or b is None or Gvbase is None
        # backward compatibility for pyscf-1.2, in which the argument Gvbase is gs
        or (Gvbase is not None and isinstance(Gvbase[0], (int, numpy.integer)))):
        p_gxyzT = lib.c_null_ptr()
        p_gs = (ctypes.c_int*3)(0,0,0)
        p_b = (ctypes.c_double*1)(0)
        eval_gz = 'GTO_Gv_general'
    else:
        if abs(b-numpy.diag(b.diagonal())).sum() < 1e-8:
            eval_gz = 'GTO_Gv_orth'
        else:
            eval_gz = 'GTO_Gv_nonorth'
        gxyzT = numpy.asarray(gxyz.T, order='C', dtype=numpy.int32)
        p_gxyzT = gxyzT.ctypes.data_as(ctypes.c_void_p)
        bqGv = numpy.hstack((b.ravel(), q) + Gvbase)
        p_b = bqGv.ctypes.data_as(ctypes.c_void_p)
        p_gs = (ctypes.c_int*3)(*[len(x) for x in Gvbase])

    ao_loc = gto.moleintor.make_loc(mol._bas, intor)

    if aosym == 's1':
        if (shls_slice[:2] == shls_slice[2:4] and
            intor.startswith('GTO_ft_ovlp')):
            fill = 'fill_s1hermi'
        else:
            fill = 'fill_s1'
        ni = int(ao_loc[shls_slice[1]] - ao_loc[shls_slice[0]])
        nj = int(ao_loc[shls_slice[3]] - ao_loc[shls_slice[2]])
        shape = (nj, ni, nGv)
    else:
        fill = 'fill_s2'
        i0 = int(ao_loc[shls_slice[0]])
        i1 = int(ao_loc[shls_slice[1]])
        nij = i1*(i1+1)//2 - i0*(i0+1)//2
        shape = (nij, nGv)
    if comp != 1:
        shape = (comp,) + shape

    if return_complex:
        fill = 'GTO_ft_z' + fill
        dtype = numpy.complex128
    else:
        if fill == 'fill_s2':
            raise NotImplementedError
        fill = 'GTO_ft_d' + fill
        dtype = numpy.double
        shape = (2,) + shape

    if out is None:
        out = numpy.zeros(shape, dtype=dtype)
    else:
        out = numpy.ndarray(shape, dtype=dtype, buffer=out)
        out[:] = 0

    if aosym == 's1':
        out = numpy.rollaxis(out, -2, -3)
        out = numpy.rollaxis(out, -1, -3)
    else:
        out = numpy.rollaxis(out, -1, -2)
    if nGv == 0:
        return out

    fn = libcgto.GTO_ft_fill_drv
    intor = getattr(libcgto, intor)
    eval_gz = getattr(libcgto, eval_gz)
    fill = getattr(libcgto, fill)
    phase = 0
    if ovlp_mask is None:
        nish = shls_slice[1] - shls_slice[0]
        njsh = shls_slice[3] - shls_slice[2]
        ovlp_mask = numpy.ones((nish,njsh), dtype=numpy.int8, order='F')
    else:
        ovlp_mask = numpy.asarray(ovlp_mask, dtype=numpy.int8, order='F')

    fn(intor, eval_gz, fill,
       out.ctypes.data_as(ctypes.c_void_p),
       ovlp_mask.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(comp), (ctypes.c_int*4)(*shls_slice),
       ao_loc.ctypes.data_as(ctypes.c_void_p), ctypes.c_double(phase),
       GvT.ctypes.data_as(ctypes.c_void_p),
       p_b, p_gxyzT, p_gs, ctypes.c_int(nGv),
       mol._atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(mol.natm),
       mol._bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(mol.nbas),
       mol._env.ctypes.data_as(ctypes.c_void_p))
    return out


# gxyz is the index for Gvbase
def ft_ao(mol, Gv, shls_slice=None, b=numpy.eye(3),
          gxyz=None, Gvbase=None, verbose=None):
    r'''Analytical FT transform AO
    \int mu(r) exp(-ikr) dr^3

    The output tensor has the shape [nGv, nao]
    '''
    if shls_slice is None:
        shls_slice = (0, mol.nbas)
    nGv = Gv.shape[0]
    if (gxyz is None or b is None or Gvbase is None
        # backward compatibility for pyscf-1.2, in which the argument Gvbase is gs
        or (Gvbase is not None and isinstance(Gvbase[0], (int, numpy.integer)))):
        GvT = numpy.asarray(Gv.T, order='C')
        p_gxyzT = lib.c_null_ptr()
        p_gs = (ctypes.c_int*3)(0,0,0)
        p_b = (ctypes.c_double*1)(0)
        eval_gz = 'GTO_Gv_general'
    else:
        if abs(b-numpy.diag(b.diagonal())).sum() < 1e-8:
            eval_gz = 'GTO_Gv_orth'
        else:
            eval_gz = 'GTO_Gv_nonorth'
        GvT = numpy.asarray(Gv.T, order='C')
        gxyzT = numpy.asarray(gxyz.T, order='C', dtype=numpy.int32)
        p_gxyzT = gxyzT.ctypes.data_as(ctypes.c_void_p)
        b = numpy.hstack((b.ravel(), numpy.zeros(3)) + Gvbase)
        p_b = b.ctypes.data_as(ctypes.c_void_p)
        p_gs = (ctypes.c_int*3)(*[len(x) for x in Gvbase])

    fn = libcgto.GTO_ft_fill_drv
    if mol.cart:
        intor = getattr(libcgto, 'GTO_ft_ovlp_cart')
    else:
        intor = getattr(libcgto, 'GTO_ft_ovlp_sph')
    eval_gz = getattr(libcgto, eval_gz)
    fill = getattr(libcgto, 'GTO_ft_zfill_s1')

    ghost_atm = numpy.array([[0,0,0,0,0,0]], dtype=numpy.int32)
    ghost_bas = numpy.array([[0,0,1,1,0,0,3,0]], dtype=numpy.int32)
    ghost_env = numpy.zeros(4)
    ghost_env[3] = numpy.sqrt(4*numpy.pi)  # s function spherical norm
    atm, bas, env = gto.conc_env(mol._atm, mol._bas, mol._env,
                                 ghost_atm, ghost_bas, ghost_env)
    ao_loc = mol.ao_loc_nr()
    nao = int(ao_loc[mol.nbas])
    ao_loc = numpy.asarray(numpy.hstack((ao_loc, [nao+1])), dtype=numpy.int32)
    ni = int(ao_loc[shls_slice[1]] - ao_loc[shls_slice[0]])
    shape = (ni, nGv)
    mat = numpy.zeros(shape, order='C', dtype=numpy.complex128)
    phase = 0
    if nGv == 0:
        return mat

    nish = shls_slice[1] - shls_slice[0]
    ovlp_mask = numpy.ones(nish, dtype=numpy.int8)

    shls_slice = shls_slice + (mol.nbas, mol.nbas+1)
    fn(intor, eval_gz, fill,
       mat.ctypes.data_as(ctypes.c_void_p),
       ovlp_mask.ctypes.data_as(ctypes.c_void_p),
       ctypes.c_int(1), (ctypes.c_int*4)(*shls_slice),
       ao_loc.ctypes.data_as(ctypes.c_void_p), ctypes.c_double(phase),
       GvT.ctypes.data_as(ctypes.c_void_p),
       p_b, p_gxyzT, p_gs, ctypes.c_int(nGv),
       atm.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(len(atm)),
       bas.ctypes.data_as(ctypes.c_void_p), ctypes.c_int(len(bas)),
       env.ctypes.data_as(ctypes.c_void_p))
    return mat.T



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#
# Ref:
# Chem Phys Lett, 256, 454
# J. Mol. Struct. THEOCHEM, 914, 3
#


import numpy
from pyscf import lib
from pyscf import symm
from pyscf.tdscf import rhf
from pyscf.tdscf._lr_eig import eigh as lr_eigh
from pyscf.dft.rks import KohnShamDFT
from pyscf import __config__


class TDA(rhf.TDA):
    def nuc_grad_method(self):
        from pyscf.grad import tdrks
        return tdrks.Gradients(self)

class TDDFT(rhf.TDHF):
    def nuc_grad_method(self):
        from pyscf.grad import tdrks
        return tdrks.Gradients(self)

RPA = TDRKS = TDDFT

class CasidaTDDFT(TDDFT, TDA):
    '''Solve the Casida TDDFT formula (A-B)(A+B)(X+Y) = (X+Y)w^2
    '''

    get_init_guess = TDA.get_init_guess
    get_precond = TDA.get_precond

    def gen_vind(self, mf=None):
        if mf is None:
            mf = self._scf
        wfnsym = self.wfnsym
        singlet = self.singlet

        mol = mf.mol
        mask = self.get_frozen_mask()
        mo_coeff = mf.mo_coeff[:, mask]
        assert mo_coeff.dtype == numpy.double
        mo_energy = mf.mo_energy[mask]
        mo_occ = mf.mo_occ[mask]
        nao, nmo = mo_coeff.shape
        occidx = numpy.where(mo_occ==2)[0]
        viridx = numpy.where(mo_occ==0)[0]
        nocc = len(occidx)
        nvir = len(viridx)
        orbv = mo_coeff[:,viridx]
        orbo = mo_coeff[:,occidx]

        if wfnsym is not None and mol.symmetry:
            if isinstance(wfnsym, str):
                wfnsym = symm.irrep_name2id(mol.groupname, wfnsym)
            wfnsym = wfnsym % 10  # convert to D2h subgroup
            sym_forbid = rhf._get_x_sym_table(self) != wfnsym

        e_ia = (mo_energy[viridx].reshape(-1,1) - mo_energy[occidx]).T
        if wfnsym is not None and mol.symmetry:
            e_ia[sym_forbid] = 0
        d_ia = numpy.sqrt(e_ia)
        ed_ia = e_ia * d_ia
        hdiag = e_ia.ravel() ** 2

        vresp = self.gen_response(singlet=singlet, hermi=1)

        def vind(zs):
            zs = numpy.asarray(zs).reshape(-1,nocc,nvir)
            # *2 for double occupancy
            dms = lib.einsum('xov,pv,qo->xpq', zs * (d_ia*2), orbv, orbo)
            # +cc for A+B and K_{ai,jb} in A == K_{ai,bj} in B
            dms = dms + dms.transpose(0,2,1)

            v1ao = vresp(dms)
            v1mo = lib.einsum('xpq,qo,pv->xov', v1ao, orbo, orbv)

            # numpy.sqrt(e_ia) * (e_ia*d_ia*z + v1mo)
            v1mo += numpy.einsum('xov,ov->xov', zs, ed_ia)
            v1mo *= d_ia
            return v1mo.reshape(v1mo.shape[0],-1)

        return vind, hdiag

    def kernel(self, x0=None, nstates=None):
        '''TDDFT diagonalization solver
        '''
        cpu0 = (lib.logger.process_clock(), lib.logger.perf_counter())
        mol = self.mol
        mf = self._scf
        if mf._numint.libxc.is_hybrid_xc(mf.xc):
            raise RuntimeError('%s cannot be used with hybrid functional'
                               % self.__class__)
        self.check_sanity()
        self.dump_flags()
        if nstates is None:
            nstates = self.nstates
        else:
            self.nstates = nstates

        log = lib.logger.Logger(self.stdout, self.verbose)

        vind, hdiag = self.gen_vind(self._scf)
        precond = self.get_precond(hdiag)

        def pickeig(w, v, nroots, envs):
            idx = numpy.where(w > self.positive_eig_threshold)[0]
            return w[idx], v[:,idx], idx

        x0sym = None
        if x0 is None:
            x0, x0sym = self.get_init_guess(
                self._scf, self.nstates, return_symmetry=True)
        elif mol.symmetry:
            x_sym = rhf._get_x_sym_table(self).ravel()
            x0sym = [rhf._guess_wfnsym_id(self, x_sym, x) for x in x0]

        self.converged, w2, x1 = lr_eigh(
            vind, x0, precond, tol_residual=self.conv_tol, lindep=self.lindep,
            nroots=nstates, x0sym=x0sym, pick=pickeig, max_cycle=self.max_cycle,
            max_memory=self.max_memory, verbose=log)

        mask = self.get_frozen_mask()
        mo_energy = self._scf.mo_energy[mask]
        mo_occ = self._scf.mo_occ[mask]
        occidx = numpy.where(mo_occ==2)[0]
        viridx = numpy.where(mo_occ==0)[0]
        e_ia = (mo_energy[viridx,None] - mo_energy[occidx]).T
        e_ia = numpy.sqrt(e_ia)
        def norm_xy(w, z):
            zp = e_ia * z.reshape(e_ia.shape)
            zm = w/e_ia * z.reshape(e_ia.shape)
            x = (zp + zm) * .5
            y = (zp - zm) * .5
            norm = lib.norm(x)**2 - lib.norm(y)**2
            if norm < 0:
                log.warn('TDDFT amplitudes |X| smaller than |Y|')
            norm = abs(.5/norm)**.5  # normalize to 0.5 for alpha spin
            return (x*norm, y*norm)

        idx = numpy.where(w2 > self.positive_eig_threshold)[0]
        self.e = numpy.sqrt(w2[idx])
        self.xy = [norm_xy(self.e[i], x1[i]) for i in idx]

        if self.chkfile:
            lib.chkfile.save(self.chkfile, 'tddft/e', self.e)
            lib.chkfile.save(self.chkfile, 'tddft/xy', self.xy)

        log.timer('TDDFT', *cpu0)
        self._finalize()
        return self.e, self.xy

    def nuc_grad_method(self):
        from pyscf.grad import tdrks
        return tdrks.Gradients(self)

TDDFTNoHybrid = CasidaTDDFT

class dRPA(TDDFTNoHybrid):
    def __init__(self, mf, frozen=None):
        if not isinstance(mf, KohnShamDFT):
            raise RuntimeError("direct RPA can only be applied with DFT; for HF+dRPA, use .xc='hf'")
        mf = mf.to_rks()
        # commit fc8d1967995b7e033b60d4428ddcca87aac78e4f handles xc='' .
        # xc='0*LDA' is equivalent to xc=''
        #mf.xc = '0.0*LDA'
        mf.xc = ''
        TDDFTNoHybrid.__init__(self, mf, frozen)

TDH = dRPA

class dTDA(TDA):
    def __init__(self, mf, frozen=None):
        if not isinstance(mf, KohnShamDFT):
            raise RuntimeError("direct TDA can only be applied with DFT; for HF+dTDA, use .xc='hf'")
        mf = mf.to_rks()
        # commit fc8d1967995b7e033b60d4428ddcca87aac78e4f handles xc='' .
        # xc='0*LDA' is equivalent to xc=''
        #mf.xc = '0.0*LDA'
        mf.xc = ''
        TDA.__init__(self, mf, frozen)


def tddft(mf, frozen=None):
    '''Driver to create TDDFT or CasidaTDDFT object'''
    if mf._numint.libxc.is_hybrid_xc(mf.xc):
        return TDDFT(mf, frozen)
    else:
        return CasidaTDDFT(mf, frozen)

from pyscf import dft
dft.rks.RKS.TDA           = dft.rks_symm.RKS.TDA           = lib.class_as_method(TDA)
dft.rks.RKS.TDHF          = dft.rks_symm.RKS.TDHF          = None
#dft.rks.RKS.TDDFT         = dft.rks_symm.RKS.TDDFT         = lib.class_as_method(TDDFT)
dft.rks.RKS.TDDFTNoHybrid = dft.rks_symm.RKS.TDDFTNoHybrid = lib.class_as_method(TDDFTNoHybrid)
dft.rks.RKS.CasidaTDDFT   = dft.rks_symm.RKS.CasidaTDDFT   = lib.class_as_method(CasidaTDDFT)
dft.rks.RKS.TDDFT         = dft.rks_symm.RKS.TDDFT         = tddft
dft.rks.RKS.dTDA          = dft.rks_symm.RKS.dTDA          = lib.class_as_method(dTDA)
dft.rks.RKS.dRPA          = dft.rks_symm.RKS.dRPA          = lib.class_as_method(dRPA)
dft.roks.ROKS.TDA           = dft.rks_symm.ROKS.TDA           = None
dft.roks.ROKS.TDHF          = dft.rks_symm.ROKS.TDHF          = None
dft.roks.ROKS.TDDFT         = dft.rks_symm.ROKS.TDDFT         = None
dft.roks.ROKS.TDDFTNoHybrid = dft.rks_symm.ROKS.TDDFTNoHybrid = None
dft.roks.ROKS.dTDA          = dft.rks_symm.ROKS.dTDA          = None
dft.roks.ROKS.dRPA          = dft.rks_symm.ROKS.dRPA          = None



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.lib import kpts as libkpts
from pyscf.pbc.scf import khf, khf_ksymm, kuhf_ksymm
from pyscf.pbc.dft import gen_grid, multigrid
from pyscf.pbc.dft import rks, krks, kuks

@lib.with_doc(kuks.get_veff.__doc__)
def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpts=None, kpts_band=None):
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpts is None: kpts = ks.kpts
    if isinstance(kpts, np.ndarray):
        return kuks.get_veff(ks, cell, dm, dm_last, vhf_last, hermi, kpts, kpts_band)

    t0 = (logger.process_clock(), logger.perf_counter())

    ground_state = kpts_band is None
    if kpts_band is None:
        kpts_band = kpts.kpts_ibz

    ni = ks._numint
    if isinstance(ni, multigrid.MultiGridNumInt):
        if ks.do_nlc():
            raise NotImplementedError(f'MultiGrid for NLC functional {ks.xc} + {ks.nlc}')
        j_in_xc = ni.xc_with_j
    else:
        ks.initialize_grids(cell, dm, kpts)
        j_in_xc = False

    max_memory = ks.max_memory - lib.current_memory()[0]
    n, exc, vxc = ni.nr_uks(cell, ks.grids, ks.xc, dm, 0, hermi,
                            kpts=kpts, kpts_band=kpts_band,
                            max_memory=max_memory)
    logger.info(ks, 'nelec by numeric integration = %s', n)
    if ks.do_nlc():
        if ni.libxc.is_nlc(ks.xc):
            xc = ks.xc
        else:
            assert ni.libxc.is_nlc(ks.nlc)
            xc = ks.nlc
        n, enlc, vnlc = ni.nr_nlc_vxc(cell, ks.nlcgrids, xc, dm[0]+dm[1],
                                      0, hermi, kpts, max_memory=max_memory)
        exc += enlc
        vxc += vnlc
        logger.info(ks, 'nelec with nlc grids = %s', n)
    t0 = logger.timer(ks, 'vxc', *t0)

    weight = kpts.weights_ibz
    vj, vk = krks._get_jk(ks, cell, dm, hermi, kpts, kpts_band, with_j=not j_in_xc)
    if j_in_xc:
        ecoul = vxc.ecoul
    else:
        vj = vj[0] + vj[1]
        vxc += vj
        ecoul = None
        if ground_state:
            ecoul = np.einsum('K,nKij,Kji->', weight, dm, vj) * .5
    if ni.libxc.is_hybrid_xc(ks.xc):
        vxc -= vk
        if ground_state:
            exc -= np.einsum('K,nKij,nKji->', weight, dm, vk).real * .5
    vxc = lib.tag_array(vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    logger.timer(ks, 'veff', *t0)
    return vxc

class KsymAdaptedKUKS(kuks.KUKS, kuhf_ksymm.KUHF):

    get_veff = get_veff

    kpts = khf_ksymm.KsymAdaptedKSCF.kpts
    get_ovlp = khf_ksymm.KsymAdaptedKSCF.get_ovlp
    get_hcore = khf_ksymm.KsymAdaptedKSCF.get_hcore
    get_jk = khf_ksymm.KsymAdaptedKSCF.get_jk
    init_guess_by_chkfile = khf_ksymm.KsymAdaptedKSCF.init_guess_by_chkfile
    dump_chk = khf_ksymm.KsymAdaptedKSCF.dump_chk

    nelec = kuhf_ksymm.KUHF.nelec
    get_init_guess = kuhf_ksymm.KUHF.get_init_guess
    get_occ = kuhf_ksymm.KUHF.get_occ
    eig = kuhf_ksymm.KUHF.eig
    get_orbsym = kuhf_ksymm.KUHF.get_orbsym
    orbsym = kuhf_ksymm.KUHF.orbsym
    _finalize = kuhf_ksymm.KUHF._finalize

    def __init__(self, cell, kpts=libkpts.KPoints(), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'),
                 **kwargs):
        kuhf_ksymm.KUHF.__init__(self, cell, kpts, exxdiv=exxdiv, **kwargs)
        rks.KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        kuhf_ksymm.KUHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    def energy_elec(self, dm_kpts=None, h1e_kpts=None, vhf=None):
        if h1e_kpts is None: h1e_kpts = self.get_hcore(self.cell, self.kpts)
        if dm_kpts is None: dm_kpts = self.make_rdm1()
        if vhf is None or getattr(vhf, 'ecoul', None) is None:
            vhf = self.get_veff(self.cell, dm_kpts)

        weight = self.kpts.weights_ibz
        e1 = np.einsum('k,kij,kji', weight, h1e_kpts, dm_kpts[0]+dm_kpts[1])
        ecoul = vhf.ecoul
        exc = vhf.exc
        tot_e = e1 + ecoul + exc
        self.scf_summary['e1'] = e1.real
        self.scf_summary['coul'] = ecoul.real
        self.scf_summary['exc'] = exc.real
        logger.debug(self, 'E1 = %s  Ecoul = %s  Exc = %s', e1, ecoul, exc)
        if khf.CHECK_COULOMB_IMAG and abs(ecoul.imag) > self.cell.precision*10:
            logger.warn(self, "Coulomb energy has imaginary part %s. "
                        "Coulomb integrals (e-e, e-N) may not converge !",
                        ecoul.imag)
        return tot_e.real, ecoul.real + exc.real

    def to_hf(self):
        '''Convert to KRHF object.'''
        from pyscf.pbc.scf.kuhf_ksymm import KUHF
        return self._transfer_attrs_(KUHF(self.cell, self.kpts))


KUKS = KsymAdaptedKUKS



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Abdelrahman Ahmed <>
#         Samragni Banerjee <samragnibanerjee4@gmail.com>
#         James Serna <jamcar456@gmail.com>
#         Terrence Stahl <>
#         Alexander Sokolov <alexander.y.sokolov@gmail.com>
#

import numpy as np

def get_ovvv_df(myadc, Lov, Lvv, p, chnk_size):

    ''' Returns approximate ovvv integrals used in restricted implementation'''

    nvir = myadc._nvir
    naux = myadc.with_df.get_naoaux()

    Lvv = Lvv.reshape(naux,nvir*nvir)
    nocc = Lov.size//(naux*nvir)
    Lov = Lov.reshape(naux,nocc,nvir)

    if chnk_size < nocc:
        Lov_temp = np.ascontiguousarray(Lov.transpose(1,2,0)[p:p+chnk_size].reshape(-1,naux))
    else:
        Lov_temp = np.ascontiguousarray(Lov.transpose(1,2,0).reshape(-1,naux))

    ovvv = np.dot(Lov_temp, Lvv)
    ovvv = ovvv.reshape(-1, nvir, nvir, nvir)
    return ovvv

def get_ovvv_spin_df(myadc, Lov, Lvv, p, chnk_size):

    ''' Returns approximate ovvv integrals (different spin cases) used in unrestricted implementation '''

    nocc = Lov.shape[1]
    nvir_1 = Lov.shape[2]
    nvir_2 = Lvv.shape[1]
    naux = myadc.with_df.get_naoaux()

    Lvv = Lvv.reshape(naux,nvir_2*nvir_2)
    Lov = Lov.reshape(naux,nocc,nvir_1)

    if chnk_size < nocc:
        #Lov_temp = np.ascontiguousarray(Lov.T[:,p:p+chnk_size,:].reshape(-1,naux))
        Lov_temp = np.ascontiguousarray(Lov.transpose(1,2,0)[p:p+chnk_size].reshape(-1,naux))
    else:
        Lov_temp = np.ascontiguousarray(Lov.transpose(1,2,0).reshape(-1,naux))

    ovvv = np.dot(Lov_temp, Lvv)
    ovvv = ovvv.reshape(-1, nvir_1, nvir_2, nvir_2)
    #vvvv = np.ascontiguousarray(vvvv.transpose(0,2,1,3)).reshape(-1, nvir, nvir * nvir)
    return ovvv


def get_vvvv_df(myadc, Lvv, p, chnk_size):

    ''' Returns approximate vvvv integrals used in restricted implementation'''

    nvir = myadc._nvir
    naux = myadc.with_df.get_naoaux()

    Lvv = Lvv.reshape(naux,nvir,nvir)

    if chnk_size < nvir:
        Lvv_temp = np.ascontiguousarray(Lvv.T[p:p+chnk_size].reshape(-1,naux))
    else:
        Lvv_temp = np.ascontiguousarray(Lvv.T.reshape(-1,naux))

    Lvv = Lvv.reshape(naux,nvir*nvir)
    vvvv = np.dot(Lvv_temp, Lvv)
    vvvv = vvvv.reshape(-1, nvir, nvir, nvir)
    vvvv = vvvv.transpose(0,2,1,3).reshape(-1, nvir, nvir, nvir)
    vvvv = np.ascontiguousarray(vvvv)

    return vvvv


def get_vvvv_antisym_df(myadc, Lvv, p, chnk_size, pack = True):

    ''' Returns approximate antisymmetrized vvvv integrals (alpha/beta spin) used in unrestricted implementation'''

    naux = myadc.with_df.get_naoaux()
    nvir = Lvv.shape[1]
    ind_vv_g = np.tril_indices(nvir, k=-1)

    if chnk_size < nvir:
        Lvv_temp = np.ascontiguousarray(Lvv.T[p:p+chnk_size].reshape(-1,naux))
    else:
        Lvv_temp = np.ascontiguousarray(Lvv.T.reshape(-1,naux))

    Lvv = Lvv.reshape(naux,nvir*nvir)
    vvvv = np.dot(Lvv_temp, Lvv)
    vvvv = vvvv.reshape(-1, nvir, nvir, nvir)
    vvvv = vvvv.transpose(0,2,1,3)
    vvvv -= vvvv.transpose(0,1,3,2)

    if pack:
        vvvv = np.ascontiguousarray(vvvv[:, :, ind_vv_g[0], ind_vv_g[1]])
    else:
        vvvv = np.ascontiguousarray(vvvv)

    return vvvv


def get_vVvV_df(myadc, Lvv, LVV, p, chnk_size):

    ''' Returns approximate vvvv integrals (mixed spin) used in unrestricted implementation'''

    naux = myadc.with_df.get_naoaux()
    nvir_1 = Lvv.shape[1]
    nvir_2 = LVV.shape[1]

    if chnk_size < nvir_1:
        Lvv_temp = np.ascontiguousarray(Lvv.T[p:p+chnk_size].reshape(-1,naux))
    else:
        Lvv_temp = np.ascontiguousarray(Lvv.T.reshape(-1,naux))

    LVV = LVV.reshape(naux,nvir_2*nvir_2)
    vvvv = np.dot(Lvv_temp, LVV)
    vvvv = vvvv.reshape(-1,nvir_1,nvir_2,nvir_2)
    vvvv = vvvv.transpose(0,2,1,3)
    vvvv = np.ascontiguousarray(vvvv)

    return vvvv



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Yang Gao <younggao1994@gmail.com>

#
'''
Analytical electron-phonon matrix for restricted kohm sham
'''

import numpy as np
from pyscf import lib
from pyscf.hessian import rks as rks_hess
from pyscf.hessian import rhf as rhf_hess
from pyscf.grad import rks as rks_grad
from pyscf.dft import numint
from pyscf.eph import rhf as rhf_eph
from pyscf.data.nist import MP_ME

CUTOFF_FREQUENCY = rhf_eph.CUTOFF_FREQUENCY
KEEP_IMAG_FREQUENCY = rhf_eph.KEEP_IMAG_FREQUENCY

def _get_vxc_deriv1(hessobj, mo_coeff, mo_occ, max_memory):
    """" This functions is slightly different from hessian.rks._get_vxc_deriv1 in that <\nabla u|Vxc|v> is removed"""
    mol = hessobj.mol
    mf = hessobj.base
    if hessobj.grids is not None:
        grids = hessobj.grids
    else:
        grids = mf.grids
    if grids.coords is None:
        grids.build(with_non0tab=True)

    nao, nmo = mo_coeff.shape
    ni = mf._numint
    xctype = ni._xc_type(mf.xc)
    aoslices = mol.aoslice_by_atom()
    shls_slice = (0, mol.nbas)
    ao_loc = mol.ao_loc_nr()
    dm0 = mf.make_rdm1(mo_coeff, mo_occ)
    vmat = np.zeros((mol.natm,3,nao,nao))
    max_memory = max(2000, max_memory-vmat.size*8/1e6)
    if xctype == 'LDA':
        ao_deriv = 1
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            rho = ni.eval_rho2(mol, ao[0], mo_coeff, mo_occ, mask, 'LDA')
            vxc, fxc = ni.eval_xc(mf.xc, rho, 0, deriv=2)[1:3]
            frr = fxc[0]
            ao_dm0 = numint._dot_ao_dm(mol, ao[0], dm0, mask, shls_slice, ao_loc)
            for ia in range(mol.natm):
                p0, p1 = aoslices[ia][2:]
                rho1 = np.einsum('xpi,pi->xp', ao[1:,:,p0:p1], ao_dm0[:,p0:p1])
                aow = np.einsum('pi,xp->xpi', ao[0], weight*frr*rho1)
                rks_grad._d1_dot_(vmat[ia], mol, aow, ao[0], mask, ao_loc, True)
            ao_dm0 = aow = None

        for ia in range(mol.natm):
            vmat[ia] = -vmat[ia] - vmat[ia].transpose(0,2,1)

    elif xctype == 'GGA':
        ao_deriv = 2
        # v_ip = np.zeros((3,nao,nao))
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            rho = ni.eval_rho2(mol, ao[:4], mo_coeff, mo_occ, mask, 'GGA')
            vxc, fxc = ni.eval_xc(mf.xc, rho, 0, deriv=2)[1:3]

            wv = numint._rks_gga_wv0(rho, vxc, weight)
            # rks_grad._gga_grad_sum_(v_ip, mol, ao, wv, mask, ao_loc)
            ao_dm0 = [numint._dot_ao_dm(mol, ao[i], dm0, mask, shls_slice, ao_loc)
                      for i in range(4)]
            for ia in range(mol.natm):
                wv = dR_rho1 = rks_hess._make_dR_rho1(ao, ao_dm0, ia, aoslices, xctype)
                wv[0] = numint._rks_gga_wv1(rho, dR_rho1[0], vxc, fxc, weight)
                wv[1] = numint._rks_gga_wv1(rho, dR_rho1[1], vxc, fxc, weight)
                wv[2] = numint._rks_gga_wv1(rho, dR_rho1[2], vxc, fxc, weight)
                aow = np.einsum('npi,Xnp->Xpi', ao[:4], wv)
                rks_grad._d1_dot_(vmat[ia], mol, aow, ao[0], mask, ao_loc, True)
            ao_dm0 = aow = None

        for ia in range(mol.natm):
            vmat[ia] = -vmat[ia] - vmat[ia].transpose(0,2,1)

    elif xctype == 'MGGA':
        raise NotImplementedError('meta-GGA')

    return vmat

def get_eph(ephobj, mo1, omega, vec, mo_rep):
    mol = ephobj.mol
    mf = ephobj.base
    ni = mf._numint
    ni.libxc.test_deriv_order(mf.xc, 2, raise_error=True)
    omg, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=mol.spin)
    hybrid = ni.libxc.is_hybrid_xc(mf.xc)

    vnuc_deriv = ephobj.vnuc_generator(mol)
    aoslices = mol.aoslice_by_atom()
    vind = rhf_eph.rhf_deriv_generator(mf, mf.mo_coeff, mf.mo_occ)
    mocc = mf.mo_coeff[:,mf.mo_occ>0]
    dm0 = np.dot(mocc, mocc.T) * 2

    natoms = mol.natm
    nao = mol.nao_nr()
    mem_now = lib.current_memory()[0]
    max_memory = max(2000, mf.max_memory*.9-mem_now)
    vxc1ao = _get_vxc_deriv1(ephobj, mf.mo_coeff, mf.mo_occ, max_memory)
    vcore = []
    for ia in range(natoms):
        h1 = vnuc_deriv(ia)
        v1 = vind(mo1[ia])
        shl0, shl1, p0, p1 = aoslices[ia]
        shls_slice = (shl0, shl1) + (0, mol.nbas)*3

        if hybrid:
            vj1, vk1 = \
                    rhf_hess._get_jk(mol, 'int2e_ip1', 3, 's2kl',
                                     ['ji->s2kl', -dm0[:,p0:p1], #vj1
                                      'li->s1kj', -dm0[:,p0:p1]], #vk1
                                     shls_slice=shls_slice)
            veff = vj1 - hyb * .5 * vk1
            if omg != 0:
                with mol.with_range_coulomb(omg):
                    vk1 = \
                        rhf_hess._get_jk(mol, 'int2e_ip1', 3, 's2kl',
                                         ['li->s1kj', -dm0[:,p0:p1]], # vk1
                                         shls_slice=shls_slice)
                veff -= (alpha-hyb) * .5 * vk1
        else:
            vj1 = rhf_hess._get_jk(mol, 'int2e_ip1', 3, 's2kl',
                                        ['ji->s2kl', -dm0[:,p0:p1]], # vj1
                                        shls_slice=shls_slice)
            veff = vj1[0]
        vtot = h1 + v1 + veff + vxc1ao[ia] + veff.transpose(0,2,1)
        vcore.append(vtot)
    vcore = np.asarray(vcore).reshape(-1,nao,nao)
    mass = mol.atom_mass_list() * MP_ME
    vec = rhf_eph._freq_mass_weighted_vec(vec, omega, mass)
    mat = np.einsum('xJ,xuv->Juv', vec, vcore, optimize=True)
    if mo_rep:
        mat = np.einsum('Juv,up,vq->Jpq', mat, mf.mo_coeff.conj(), mf.mo_coeff, optimize=True)
    return mat


class EPH(rks_hess.Hessian):
    '''EPH for restricted DFT

    Attributes:
        cutoff_frequency : float or int
            cutoff frequency in cm-1. Default is 80
        keep_imag_frequency : bool
            Whether to keep imaginary frequencies in the output.  Default is False

    Saved results

        omega : numpy.ndarray
            Vibrational frequencies in au.
        vec : numpy.ndarray
            Polarization vectors of the vibration modes
        eph : numpy.ndarray
            Electron phonon matrix eph[j,a,b] (j in nmodes, a,b in norbs)
    '''
    def __init__(self, scf_method, cutoff_frequency=CUTOFF_FREQUENCY,
                 keep_imag_frequency=KEEP_IMAG_FREQUENCY):
        rks_hess.Hessian.__init__(self, scf_method)
        self.cutoff_frequency = cutoff_frequency
        self.keep_imag_frequency = keep_imag_frequency

    get_mode = rhf_eph.get_mode
    get_eph = get_eph
    vnuc_generator = rhf_eph.vnuc_generator
    kernel = rhf_eph.kernel

if __name__ == '__main__':
    from pyscf import gto, dft
    mol = gto.M()
    mol.atom = [['O', [0.000000000000, 0.000000002577,0.868557119905]],
                ['H', [0.000000000000,-1.456050381698,2.152719488376]],
                ['H', [0.000000000000, 1.456050379121,2.152719486067]]]

    mol.unit = 'Bohr'
    mol.basis = 'sto3g'
    mol.verbose=4
    mol.build() # this is a pre-computed relaxed geometry

    mf = dft.RKS(mol)
    mf.grids.level=6
    mf.xc = 'b3lyp'
    mf.conv_tol = 1e-16
    mf.conv_tol_grad = 1e-10
    mf.kernel()

    grad = mf.nuc_grad_method().kernel()
    print("Force on the atoms/au:")
    print(grad)


    myeph = EPH(mf)
    eph, omega = myeph.kernel(mo_rep=True)
    print(np.amax(eph))



#!/usr/bin/env python
# Copyright 2017-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: James D. McClain
#          Timothy Berkelbach <tim.berkelbach@gmail.com>
#

#import numpy as np
from itertools import product
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.lib import kpts_helper
import numpy
from pyscf.lib.parameters import LARGE_DENOM  # noqa
from pyscf.pbc.mp.kmp2 import (get_frozen_mask, get_nocc, get_nmo,
                               padded_mo_coeff, padding_k_idx)  # noqa
from pyscf.pbc.cc.kccsd_rhf import _get_epq
from pyscf.pbc.cc.kccsd_t_rhf import _get_epqr

#einsum = numpy.einsum
einsum = lib.einsum

#################################################
# FOLLOWING:                                    #
# J. Gauss and J. F. Stanton,                   #
# J. Chem. Phys. 103, 3561 (1995) Table III     #
#################################################

### Section (a)

def make_tau(cc, t2, t1, t1p, kconserv, fac=1., out=None):
    nkpts, nocc, nvir = t1.shape
    tau1 = numpy.ndarray(t2.shape, dtype=t2.dtype, buffer=out)
    tau1[:] = t2
    for ki in range(nkpts):
        for ka in range(nkpts):
            for kj in range(nkpts):
                kb = kconserv[ki,ka,kj]
                tmp = numpy.zeros((nocc,nocc,nvir,nvir),dtype=t2.dtype)
                if ki == ka and kj == kb:
                    tmp += einsum('ia,jb->ijab',t1[ki],t1p[kj])
                if ki == kb and kj == ka:
                    tmp -= einsum('ib,ja->ijab',t1[ki],t1p[kj])
                if kj == ka and ki == kb:
                    tmp -= einsum('ja,ib->ijab',t1[kj],t1p[ki])
                if kj == kb and ki == ka:
                    tmp += einsum('jb,ia->ijab',t1[kj],t1p[ki])
                tau1[ki,kj,ka] += fac*0.5*tmp
    return tau1

def cc_Fvv(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    fov = eris.fock[:,:nocc,nocc:].copy()
    fvv = eris.fock[:,nocc:,nocc:].copy()
    # <o(k1)v(k2)||v(k3)v(k4)> = <v(k2)o(k1)||v(k4)v(k3)> = -<v(k2)o(k1)||v(k3)v(k4)>
    eris_vovv = -eris.ovvv.transpose(1,0,2,4,3,5,6)
    tau_tilde = make_tau(cc,t2,t1,t1,kconserv,fac=0.5)
    Fae = numpy.zeros(fvv.shape, t1.dtype)
    #kconserv = kpts_helper.get_kconserv(cc._scf.cell, cc.kpts)
    for ka in range(nkpts):
        Fae[ka] += fvv[ka]
        Fae[ka] += -0.5*einsum('me,ma->ae',fov[ka],t1[ka])
        for km in range(nkpts):
            Fae[ka] += einsum('mf,amef->ae',t1[km],eris_vovv[ka,km,ka])
            for kn in range(nkpts):
                #kb = kconserv[km,ka,kn]
                Fae[ka] += -0.5*einsum('mnaf,mnef->ae',tau_tilde[km,kn,ka],
                                       eris.oovv[km,kn,ka])
    return Fae

def cc_Foo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    fov = eris.fock[:,:nocc,nocc:].copy()
    foo = eris.fock[:,:nocc,:nocc].copy()
    tau_tilde = make_tau(cc,t2,t1,t1,kconserv,fac=0.5)
    Fmi = numpy.zeros(foo.shape, t1.dtype)
    for km in range(nkpts):
        Fmi[km] += foo[km]
        Fmi[km] += 0.5*einsum('me,ie->mi',fov[km],t1[km])
        for kn in range(nkpts):
            Fmi[km] += einsum('ne,mnie->mi',t1[kn],eris.ooov[km,kn,km])
            for ke in range(nkpts):
                Fmi[km] += 0.5*einsum('inef,mnef->mi',tau_tilde[km,kn,ke],
                                      eris.oovv[km,kn,ke])
    return Fmi

def cc_Fov(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    fov = eris.fock[:,:nocc,nocc:].copy()
    Fme = numpy.zeros(fov.shape, t1.dtype)
    for km in range(nkpts):
        Fme[km] += fov[km]
        for kf in range(nkpts):
            kn = kf
            Fme[km] -= einsum('nf,mnfe->me',t1[kf],eris.oovv[km,kn,kf])
    return Fme

def cc_Woooo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    tau = make_tau(cc,t2,t1,t1,kconserv)
    Wmnij = eris.oooo.copy()
    for km in range(nkpts):
        for kn in range(nkpts):
            # Since it's not enough just to switch i and j and need to create the k_i and k_j
            # so that P(ij) switches both i,j and k_i,k_j
            #   t1[ k_j, j, e ] * v[ k_m, k_n, k_i, m, n, i, e ] -> tmp[ k_i, k_j, m, n, i, j ]
            # Here, x = k_j and y = k_i
            tmp = einsum('xje,ymnie->yxmnij',t1,eris.ooov[km,kn])
            tmp = tmp - tmp.transpose(1,0,2,3,5,4)

            ki = numpy.arange(nkpts)
            kj = kconserv[km,ki,kn]
            kij = (ki,kj)
            Wmnij[km,kn,:] += 0.25*einsum('yxijef,xmnef->ymnij',tau[kij],eris.oovv[km,kn])

            for ki in range(nkpts):
                kj = kconserv[km,ki,kn]
                Wmnij[km,kn,ki] += tmp[ki,kj]
    return Wmnij

def cc_Wvvvv(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    eris_vovv = - eris.ovvv.transpose(1,0,2,4,3,5,6)
    tau = make_tau(cc,t2,t1,t1,kconserv)
    Wabef = eris.vvvv.copy()
    for ka in range(nkpts):
        for kb in range(nkpts):

            km = numpy.arange(nkpts).tolist()
            kn = kconserv[ka,km,kb].tolist()
            kmn = (km,kn)
            Wabef[ka,kb] += 0.25*einsum('xmnab,xymnef->yabef',tau.transpose(2,0,1,3,4,5,6)[ka][kmn],eris.oovv[kmn])

            for ke in range(nkpts):
                km = kb
                tmp  = einsum('mb,amef->abef',t1[kb],eris_vovv[ka,km,ke])
                km = ka
                tmp -= einsum('ma,bmef->abef',t1[ka],eris_vovv[kb,km,ke])
                Wabef[ka,kb,ke] += -tmp
                # km + kn - ka = kb
                # => kn = ka - km + kb

    return Wabef

def cc_Wovvo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    eris_ovvo = numpy.zeros(shape=(nkpts,nkpts,nkpts,nocc,nvir,nvir,nocc),dtype=t2.dtype)
    eris_oovo = numpy.zeros(shape=(nkpts,nkpts,nkpts,nocc,nocc,nvir,nocc),dtype=t2.dtype)
    for km in range(nkpts):
        for kb in range(nkpts):
            for ke in range(nkpts):
                kj = kconserv[km,ke,kb]
                # <mb||je> -> -<mb||ej>
                eris_ovvo[km,kb,ke] = -eris.ovov[km,kb,kj].transpose(0,1,3,2)
                # <mn||je> -> -<mn||ej>
                # let kb = kn as a dummy variable
                eris_oovo[km,kb,ke] = -eris.ooov[km,kb,kj].transpose(0,1,3,2)
    Wmbej = eris_ovvo.copy()
    for km in range(nkpts):
        for kb in range(nkpts):
            for ke in range(nkpts):
                kj = kconserv[km,ke,kb]
                Wmbej[km,kb,ke] += einsum('jf,mbef->mbej',t1[kj,:,:],eris.ovvv[km,kb,ke])
                Wmbej[km,kb,ke] += -einsum('nb,mnej->mbej',t1[kb,:,:],eris_oovo[km,kb,ke])
                temp = numpy.zeros([nkpts, nocc, nocc, nvir, nvir], dtype=t2.dtype)
                for kn in range(nkpts):
                    kf = kconserv[km,ke,kn]
                    temp[kn] = -0.5*t2[kj,kn,kf].copy()
                    if kn == kb and kf == kj:
                        temp[kn] -= einsum('jf,nb->jnfb', t1[kj], t1[kn])
                Wmbej[km,kb,ke] += einsum('xjnfb, xmnef->mbej', temp, eris.oovv[km,:,ke])
    return Wmbej

def cc_Wovvo_jk(cc, t1, t2, eris, kconserv):
    nkpts, nocc, nvir = t1.shape
    Wmbej = eris.ovvo.copy()
    for km in range(nkpts):
        for kb in range(nkpts):
            for ke in range(nkpts):
                kj = kconserv[km,ke,kb]
                Wmbej[km,kb,ke] += einsum('jf,mbef->mbej',t1[kj,:,:],eris.ovvv[km,kb,ke])
                Wmbej[km,kb,ke] += -einsum('nb,mnej->mbej',t1[kb,:,:],eris.oovo[km,kb,ke])
                for kn in range(nkpts):
                    kf = kconserv[km,ke,kn]
                    Wmbej[km,kb,ke] += -0.5*einsum('jnfb,mnef->mbej',t2[kj,kn,kf],
                                                   eris.oovv[km,kn,ke])
                    if kn == kb and kf == kj:
                        Wmbej[km,kb,ke] += -einsum('jf,nb,mnef->mbej',t1[kj],t1[kn],
                                                   eris.oovv[km,kn,ke])
    return Wmbej

### Section (b)

def Fvv(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    ccFov = cc_Fov(cc,t1,t2,eris,kconserv)
    Fae = cc_Fvv(cc,t1,t2,eris,kconserv)
    for km in range(nkpts):
        Fae[km] -= 0.5*einsum('ma,me->ae', t1[km], ccFov[km])
    return Fae

def Foo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    ccFov = cc_Fov(cc,t1,t2,eris,kconserv)
    Fmi = cc_Foo(cc,t1,t2,eris,kconserv)
    for km in range(nkpts):
        Fmi[km] += 0.5*einsum('ie,me->mi',t1[km],ccFov[km])
    return Fmi

def Fov(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    Fme = cc_Fov(cc,t1,t2,eris,kconserv)
    return Fme

def Woooo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    tau = make_tau(cc,t2,t1,t1,kconserv)
    Wmnij = cc_Woooo(cc,t1,t2,eris,kconserv)
    for km in range(nkpts):
        for kn in range(nkpts):
            for ki in range(nkpts):
                kj = kconserv[km ,ki, kn]
                Wmnij[km, kn, ki] += 0.25*einsum('xijef,xmnef->mnij',tau[ki, kj, :],
                                                 eris.oovv[km, kn, :])
    return Wmnij

def Wvvvv(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    tau = make_tau(cc,t2,t1,t1,kconserv)
    Wabef = cc_Wvvvv(cc,t1,t2,eris,kconserv)
    for ka, kb, ke in kpts_helper.loop_kkk(nkpts):
        #kf = kconserv[ka, ke, kb]
        for km in range(nkpts):
            kn = kconserv[ka, km, kb]
            Wabef[ka, kb, ke] += 0.25*einsum('mnab,mnef->abef',tau[km, kn, ka],
                                             eris.oovv[km, kn, ke])
    return Wabef

def Wovvo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    Wmbej = cc_Wovvo(cc,t1,t2,eris,kconserv)
    for km, kb, ke in kpts_helper.loop_kkk(nkpts):
        kj = kconserv[km, ke, kb]
        for kn in range(nkpts):
            kf = kconserv[km, ke, kn]
            Wmbej[km, kb, ke] -= 0.5*einsum('jnfb,mnef->mbej',t2[kj, kn, kf],
                                            eris.oovv[km, kn, ke])
    return Wmbej

# Indices in the following can be safely permuted.

def Wooov(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    Wmnie = eris.ooov.copy()
    for km, kn, ki in kpts_helper.loop_kkk(nkpts):
        kf = ki
        Wmnie[km, kn, ki] += einsum('if,mnfe->mnie',t1[ki], eris.oovv[km, kn, kf])
    return Wmnie

def Wvovv(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape
    Wamef = numpy.empty((nkpts, nkpts, nkpts, nvir, nocc, nvir, nvir), dtype=eris.ovvv.dtype)
    for ka, km, ke in kpts_helper.loop_kkk(nkpts):
        kn = ka
        Wamef[ka, km, ke] = -eris.ovvv[km, ka, ke].transpose(1, 0, 2, 3)
        Wamef[ka, km, ke] -= einsum('na,nmef->amef',t1[kn],eris.oovv[kn, km, ke])
    return Wamef

def Wovoo(cc,t1,t2,eris,kconserv):
    nkpts, nocc, nvir = t1.shape

    Wmbij = eris.ovoo.copy()
    for km, kb, ki in kpts_helper.loop_kkk(nkpts):
        kj = kconserv[km, ki, kb]
        for kn in range(nkpts):
            Wmbij[km, kb, ki] += einsum('mnie,jnbe->mbij', eris.ooov[km, kn, ki], t2[kj, kn, kb])

        Wmbij[km, kb, ki] += einsum('ie,mbej->mbij', t1[ki], -eris.ovov[km, kb, kj].transpose(0, 1, 3, 2))
        for kf in range(nkpts):
            kn = kconserv[kb, kj, kf]
            Wmbij[km, kb, ki] -= einsum('ie,njbf,mnef->mbij', t1[ki], t2[kn, kj, kb], eris.oovv[km, kn, ki])
        # P(ij)
        for kn in range(nkpts):
            Wmbij[km, kb, ki] -= einsum('mnje,inbe->mbij', eris.ooov[km, kn, kj], t2[ki, kn, kb])

        Wmbij[km, kb, ki] -= einsum('je,mbei->mbij', t1[kj], -eris.ovov[km, kb, ki].transpose(0, 1, 3, 2))
        for kf in range(nkpts):
            kn = kconserv[kb, ki, kf]
            Wmbij[km, kb, ki] += einsum('je,nibf,mnef->mbij', t1[kj], t2[kn, ki, kb], eris.oovv[km, kn, kj])

    FFov = Fov(cc,t1,t2,eris,kconserv)
    WWoooo = Woooo(cc,t1,t2,eris,kconserv)
    tau = make_tau(cc,t2,t1,t1,kconserv)
    for km, kb, ki in kpts_helper.loop_kkk(nkpts):
        kj = kconserv[km, ki, kb]
        Wmbij[km, kb, ki] -= einsum('me,ijbe->mbij', FFov[km], t2[ki, kj, kb])
        Wmbij[km, kb, ki] -= einsum('nb,mnij->mbij', t1[kb], WWoooo[km, kb, ki])

    for km, kb, ki in kpts_helper.loop_kkk(nkpts):
        kj = kconserv[km, ki, kb]
        Wmbij[km, kb, ki] += 0.5 * einsum('xmbef,xijef->mbij', eris.ovvv[km, kb, :], tau[ki, kj, :])

    return Wmbij

def Wvvvo(cc,t1,t2,eris,kconserv,WWvvvv=None):
    nkpts, nocc, nvir = t1.shape
    FFov = Fov(cc,t1,t2,eris,kconserv)
    if WWvvvv is None:
        WWvvvv = Wvvvv(cc,t1,t2,eris,kconserv)

    eris_ovvo = numpy.zeros(shape=(nkpts,nkpts,nkpts,nocc,nvir,nvir,nocc),dtype=t2.dtype)
    for km in range(nkpts):
        for kb in range(nkpts):
            for ke in range(nkpts):
                kj = kconserv[km,ke,kb]
                eris_ovvo[km,kb,ke] = -eris.ovov[km,kb,kj].transpose(0,1,3,2)

    tmp1 = numpy.zeros((nkpts, nkpts, nkpts, nvir, nvir, nvir, nocc),dtype=t2.dtype)
    tmp2 = numpy.zeros((nkpts, nkpts, nkpts, nvir, nvir, nvir, nocc),dtype=t2.dtype)
    for ka, kb, ke in kpts_helper.loop_kkk(nkpts):
        ki = kconserv[ka,ke,kb]
        tmp2[ka,kb,ke] += einsum('ma,mbei->abei',t1[ka],eris_ovvo[ka,kb,ke])
        for kn in range(nkpts):
            tmp2[ka,kb,ke] -= einsum('ma,nibf,mnef->abei',t1[ka],t2[kn,ki,kb],eris.oovv[ka,kn,ke])
        for km in range(nkpts):
            tmp1[ka,kb,ke] += einsum('mbef,miaf->abei',eris.ovvv[km,kb,ke],t2[km,ki,ka])
    tau = make_tau(cc,t2,t1,t1,kconserv)

    Wabei = -tmp1 + tmp1.transpose(1,0,2,4,3,5,6)
    Wabei -= tmp2 - tmp2.transpose(1,0,2,4,3,5,6)
    for ka, kb, ke in kpts_helper.loop_kkk(nkpts):
        ki = kconserv[ka, ke, kb]
        Wabei[ka, kb, ke] += eris.ovvv[ki, ke, kb].conj().transpose(3, 2, 1, 0)
        Wabei[ka, kb, ke] += einsum('if,abef->abei',t1[ki],WWvvvv[ka, kb, ke])
        Wabei[ka, kb, ke] -= einsum('me,miab->abei',FFov[ke],t2[ke, ki, ka])
        for km in range(nkpts):
            kn = kconserv[ka, km, kb]
            Wabei[ka, kb, ke] += 0.5 * einsum('nmie,mnab->abei',
                                              eris.ooov[kn, km, ki],
                                              tau[km, kn, ka])
    return Wabei

def get_full_t3p2(mycc, t1, t2, eris):
    '''Build the entire T3[2] array in memory.
    '''
    nkpts = mycc.nkpts
    nocc = mycc.nocc
    nmo = mycc.nmo
    nvir = nmo - nocc
    kconserv = mycc.khelper.kconserv

    def get_wijkabc(ki, kj, kk, ka, kb, kc):
        '''Build T3[2] for `ijkabc` at a given set of k-points'''
        km = kconserv[kc, kk, kb]
        kf = kconserv[kk, kc, kj]
        ret = einsum('kjcf,ifab->ijkabc', t2[kk,kj,kc], eris.ovvv[ki,kf,ka].conj())
        ret = ret - einsum('jima,mkbc->ijkabc', eris.ooov[kj,ki,km].conj(), t2[km,kk,kb])
        return ret

    #fock = eris.fock
    #fov = fock[:, :nocc, nocc:]
    #foo = numpy.array([fock[ikpt, :nocc, :nocc].diagonal() for ikpt in range(nkpts)])
    #fvv = numpy.array([fock[ikpt, nocc:, nocc:].diagonal() for ikpt in range(nkpts)])
    mo_energy_occ = numpy.array([eris.mo_energy[ki][:nocc] for ki in range(nkpts)])
    mo_energy_vir = numpy.array([eris.mo_energy[ki][nocc:] for ki in range(nkpts)])

    mo_e_o = mo_energy_occ
    mo_e_v = mo_energy_vir

    # Get location of padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = padding_k_idx(mycc, kind="split")

    t3 = numpy.empty((nkpts,nkpts,nkpts,nkpts,nkpts,nocc,nocc,nocc,nvir,nvir,nvir),
                      dtype=t2.dtype)
    for ki, kj, kk, ka, kb in product(range(nkpts), repeat=5):
        kc = kpts_helper.get_kconserv3(mycc._scf.cell, mycc.kpts,
                                       [ki, kj, kk, ka, kb])
        # Perform P(abc)
        t3[ki,kj,kk,ka,kb] =  get_wijkabc(ki,kj,kk,ka,kb,kc)
        t3[ki,kj,kk,ka,kb] += get_wijkabc(ki,kj,kk,kb,kc,ka).transpose(0,1,2,5,3,4)
        t3[ki,kj,kk,ka,kb] += get_wijkabc(ki,kj,kk,kc,ka,kb).transpose(0,1,2,4,5,3)

    # Perform P(ijk)
    t3 = (t3.transpose(0,1,2,3,4,5,6,7,8,9,10) +
          t3.transpose(1,2,0,3,4,6,7,5,8,9,10) +
          t3.transpose(2,0,1,3,4,7,5,6,8,9,10))

    for ki, kj, kk in product(range(nkpts), repeat=3):
        eijk = _get_epqr([0,nocc,ki,mo_e_o,nonzero_opadding],
                         [0,nocc,kj,mo_e_o,nonzero_opadding],
                         [0,nocc,kk,mo_e_o,nonzero_opadding])
        for ka, kb in product(range(nkpts), repeat=2):
            kc = kpts_helper.get_kconserv3(mycc._scf.cell, mycc.kpts,
                                           [ki, kj, kk, ka, kb])
            eabc = _get_epqr([0,nvir,ka,mo_e_v,nonzero_vpadding],
                             [0,nvir,kb,mo_e_v,nonzero_vpadding],
                             [0,nvir,kc,mo_e_v,nonzero_vpadding],
                             fac=[-1.,-1.,-1.])
            eijkabc = eijk[:, :, :, None, None, None] + eabc[None, None, None, :, :, :]
            t3[ki,kj,kk,ka,kb] /= eijkabc

    return t3

def get_t3p2_imds_slow(cc, t1, t2, eris=None, t3p2_ip_out=None, t3p2_ea_out=None):
    """Calculates T1, T2 amplitudes corrected by second-order T3 contribution
    and intermediates used in IP/EA-CCSD(T)a

    Args:
        cc (:obj:`KGCCSD`):
            Object containing coupled-cluster results.
        t1 (:obj:`ndarray`):
            T1 amplitudes.
        t2 (:obj:`ndarray`):
            T2 amplitudes from which the T3[2] amplitudes are formed.
        eris (:obj:`_PhysicistsERIs`):
            Antisymmetrized electron-repulsion integrals in physicist's notation.
        t3p2_ip_out (:obj:`ndarray`):
            Store results of the intermediate used in IP-EOM-CCSD(T)a.
        t3p2_ea_out (:obj:`ndarray`):
            Store results of the intermediate used in EA-EOM-CCSD(T)a.

    Returns:
        delta_ccsd (float):
            Difference of perturbed and unperturbed CCSD ground-state energy,
                energy(T1 + T1[2], T2 + T2[2]) - energy(T1, T2)
        pt1 (:obj:`ndarray`):
            Perturbatively corrected T1 amplitudes.
        pt2 (:obj:`ndarray`):
            Perturbatively corrected T2 amplitudes.

    Reference:
        D. A. Matthews, J. F. Stanton "A new approach to approximate..."
            JCP 145, 124102 (2016); DOI:10.1063/1.4962910, Equation 14
        Shavitt and Bartlett "Many-body Methods in Physics and Chemistry"
            2009, Equation 10.33
    """
    if eris is None:
        eris = cc.ao2mo()
    fock = eris.fock
    nkpts, nocc, nvir = t1.shape
    kconserv = cc.khelper.kconserv

    fov = [fock[ikpt, :nocc, nocc:] for ikpt in range(nkpts)]
    #foo = [fock[ikpt, :nocc, :nocc].diagonal() for ikpt in range(nkpts)]
    #fvv = [fock[ikpt, nocc:, nocc:].diagonal() for ikpt in range(nkpts)]
    mo_energy_occ = numpy.array([eris.mo_energy[ki][:nocc] for ki in range(nkpts)])
    mo_energy_vir = numpy.array([eris.mo_energy[ki][nocc:] for ki in range(nkpts)])

    # Get location of padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = padding_k_idx(cc, kind="split")

    mo_e_o = mo_energy_occ
    mo_e_v = mo_energy_vir

    ccsd_energy = cc.energy(t1, t2, eris)
    dtype = numpy.result_type(t1, t2)

    if t3p2_ip_out is None:
        t3p2_ip_out = numpy.zeros((nkpts,nkpts,nkpts,nocc,nvir,nocc,nocc), dtype=dtype)
    Wmcik = t3p2_ip_out

    if t3p2_ea_out is None:
        t3p2_ea_out = numpy.zeros((nkpts,nkpts,nkpts,nvir,nvir,nvir,nocc), dtype=dtype)
    Wacek = t3p2_ea_out

    t3 = get_full_t3p2(cc, t1, t2, eris)

    pt1 = numpy.zeros((nkpts, nocc, nvir), dtype=dtype)
    for ki in range(nkpts):
        ka = ki
        for km, kn, ke in product(range(nkpts), repeat=3):
            pt1[ki] += 0.25 * lib.einsum('mnef,imnaef->ia', eris.oovv[km,kn,ke], t3[ki,km,kn,ka,ke])
        eia = _get_epq([0,nocc,ki,mo_e_o,nonzero_opadding],
                       [0,nvir,ka,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        pt1[ki] /= eia

    pt2 = numpy.zeros((nkpts, nkpts, nkpts, nocc, nocc, nvir, nvir), dtype=dtype)
    for ki, kj, ka in product(range(nkpts), repeat=3):
        kb = kconserv[ki,ka,kj]
        for km in range(nkpts):
            pt2[ki,kj,ka] += lib.einsum('ijmabe,me->ijab', t3[ki,kj,km,ka,kb], fov[km])
            for ke in range(nkpts):
                kf = kconserv[km,ke,kb]
                pt2[ki,kj,ka] += 0.5 * lib.einsum('ijmaef,mbfe->ijab', t3[ki,kj,km,ka,ke], eris.ovvv[km,kb,kf])
                kf = kconserv[km,ke,ka]
                pt2[ki,kj,ka] -= 0.5 * lib.einsum('ijmbef,mafe->ijab', t3[ki,kj,km,kb,ke], eris.ovvv[km,ka,kf])

            for kn in range(nkpts):
                pt2[ki,kj,ka] -= 0.5 * lib.einsum('inmabe,nmje->ijab', t3[ki,kn,km,ka,kb], eris.ooov[kn,km,kj])
                pt2[ki,kj,ka] += 0.5 * lib.einsum('jnmabe,nmie->ijab', t3[kj,kn,km,ka,kb], eris.ooov[kn,km,ki])

        eia = _get_epq([0,nocc,ki,mo_e_o,nonzero_opadding],
                       [0,nvir,ka,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        ejb = _get_epq([0,nocc,kj,mo_e_o,nonzero_opadding],
                       [0,nvir,kb,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        eijab = eia[:, None, :, None] + ejb[:, None, :]
        pt2[ki,kj,ka] /= eijab

    pt1 += t1
    pt2 += t2

    for ki, kj, kk, ka, kb in product(range(nkpts), repeat=5):
        kc = kpts_helper.get_kconserv3(cc._scf.cell, cc.kpts,
                                       [ki, kj, kk, ka, kb])
        tmp = t3[ki,kj,kk,ka,kb]
        km = kconserv[ki,kc,kk]
        ke = kconserv[ka,kk,kc]

        Wmcik[km,kc,ki] += 0.5*lib.einsum('ijkabc,mjab->mcik', tmp, eris.oovv[km,kj,ka])
        Wacek[ka,kc,ke] += -0.5*lib.einsum('ijkabc,ijeb->acek', tmp, eris.oovv[ki,kj,ke])

    delta_ccsd_energy = cc.energy(pt1, pt2, eris) - ccsd_energy
    logger.info(cc, 'CCSD energy T3[2] correction : %14.8e', delta_ccsd_energy)
    return delta_ccsd_energy, pt1, pt2, Wmcik, Wacek
